# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import functools
import inspect

from ... import utils as ut
from ...localize import _
from ..rawtrigger import EncodePlayer
from ..variable import EUDVariable, IsEUDVariable
from ..variable.evcommon import _ev
from .eudtypedfuncn import EUDFullFuncN, EUDTypedFuncN, EUDXTypedFuncN, _apply_types


def EUDTypedFunc(argtypes, rettypes=None, *, traced=False):  # noqa: N802
    def _eud_typed_func(fdecl_func):
        argspec = inspect.getfullargspec(fdecl_func)
        argn = len(argspec[0])
        ut.ep_assert(
            argspec[1] is None,
            _("No variadic arguments (*args) allowed for EUDFunc."),
        )
        ut.ep_assert(
            argspec[2] is None,
            _("No variadic keyword arguments (**kwargs) allowed for EUDFunc."),
        )

        def caller(*args):
            # Cast arguments to argtypes before callee code.
            args = _apply_types(argtypes, args)
            return fdecl_func(*args)

        ret = EUDTypedFuncN(
            argn, caller, fdecl_func, argtypes, rettypes, traced=traced
        )
        functools.update_wrapper(ret, fdecl_func)
        return ret

    return _eud_typed_func


def EUDTracedTypedFunc(argtypes, rettypes=None):  # noqa: N802
    return EUDTypedFunc(argtypes, rettypes, traced=True)


def EUDFunc(fdecl_func):  # noqa: N802
    return EUDTypedFunc(None, None, traced=False)(fdecl_func)


def EUDTracedFunc(fdecl_func):  # noqa: N802
    return EUDTypedFunc(None, None, traced=True)(fdecl_func)


def EUDXTypedFunc(argmasks, argtypes, rettypes=None, *, traced=False):  # noqa: N802
    def _eudx_typed_func(fdecl_func):
        argspec = inspect.getfullargspec(fdecl_func)
        argn = len(argspec[0])
        ut.ep_assert(
            argspec[1] is None,
            _("No variadic arguments (*args) allowed for EUDFunc."),
        )
        ut.ep_assert(
            argspec[2] is None,
            _("No variadic keyword arguments (**kwargs) allowed for EUDFunc."),
        )

        def caller(*args):
            # Cast arguments to argtypes before callee code.
            args = _apply_types(argtypes, args)
            return fdecl_func(*args)

        ret = EUDXTypedFuncN(
            argn, caller, fdecl_func, argtypes, rettypes, argmasks, traced=traced
        )
        functools.update_wrapper(ret, fdecl_func)
        return ret

    return _eudx_typed_func


def EUDFullFunc(arginitvals, argtypes, rettypes=None, *, traced=False):  # noqa: N802
    def _eud_full_func(fdecl_func):
        argspec = inspect.getfullargspec(fdecl_func)
        argn = len(argspec[0])
        ut.ep_assert(
            argn >= len(arginitvals),
            _("Different number of variables({}) from initial values({})").format(
                argn, len(arginitvals)
            ),
        )
        ut.ep_assert(
            argspec[1] is None,
            _("No variadic arguments (*args) allowed for EUDFunc."),
        )
        ut.ep_assert(
            argspec[2] is None,
            _("No variadic keyword arguments (**kwargs) allowed for EUDFunc."),
        )

        def caller(*args):
            # Cast arguments to argtypes before callee code.
            args = _apply_types(argtypes, args)
            return fdecl_func(*args)

        ret = EUDFullFuncN(
            argn, arginitvals, caller, fdecl_func, argtypes, rettypes, traced=traced
        )
        functools.update_wrapper(ret, fdecl_func)
        return ret

    return _eud_full_func


def _EUDPredefineParam(*args):  # noqa: N802
    """
    Use with cautions!
    1. Don't initialize value!
    2. Reset modifier to `SetTo` when you're done!
    3. Always SetDest when assign to other variables!
    4. No EUDFunc call in function body!
    """
    fnargs, slicer = list(), list()
    for arg in args:
        if EncodePlayer(arg) == 13:  # CurrentPlayer
            fnargs.append(ut.EPD(0x6509B0))
        elif isinstance(arg, (list, tuple)):  # noqa: UP038
            fnargs.extend(arg)
        else:
            slicer.append(arg)
    while slicer:
        try:
            arg = ut.FlattenList(_ev[slice(*slicer)])
            if arg:
                fnargs.extend(arg)
                break
        except IndexError:
            _ev.append(EUDVariable())
        else:
            _ev.append(EUDVariable())

    def wrapper(f):
        f._fargs = fnargs
        f._argn = len(fnargs)
        return f

    return wrapper


def _EUDPredefineReturn(*frets):  # noqa: N802
    """
    Use with cautions!
    1. Always initialize value!
    2. Reset modifier to `SetTo` when you're done!
    3. Don't modify Dest in function body!
    4. No EUDFunc call in function body!
    """
    ut.ep_assert(frets, "empty function return")
    if len(frets) <= 2 and all(isinstance(ret, int) for ret in frets):
        while len(_ev) < max(frets):
            _ev.append(EUDVariable())
        rets = _ev[slice(*frets)]

    elif all(IsEUDVariable(ret) for ret in frets):
        rets = frets

    else:
        raise ut.EPError(_("Invalid return variable: {}").format(frets))

    rets = ut.FlattenList(rets)

    def wrapper(f):
        f._frets = rets
        f._retn = len(rets)
        return f

    return wrapper



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ... import core as c


@c.EUDFunc
def f_sqrt(n):
    from ..utilf.binsearch import EUDBinaryMax

    return EUDBinaryMax(lambda x: x * x <= n, 0, 0xFFFF)



# Copyright 2019 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import random
from typing import Any

from ..core.rawtrigger import (
    Action,
    Add,
    EncodeModifier,
    SetDeaths,
)
from ..utils import EPD
from . import modcurpl as cp

_seed = 0
_stack = []


def push() -> None:
    _stack.append(_seed)


def pop() -> None:
    _stack.pop()


def load() -> None:
    global _seed
    _seed = _stack[-1]


def rand(dest) -> tuple[Any, Any]:
    unit = random.randint(234, 65535)
    cpo = EPD(dest) - 12 * unit - _seed
    return cpo, unit


def srand() -> Action:
    r = random.randint(27, 0xFFFFFFFF)
    _loc = random.randint(1, 0xFFFFFFFF)
    u = random.randint(234, 65535)
    epd = EPD(0x6509B0) - 12 * u
    global _seed
    _seed = r
    return Action(_loc, 0, 0, 0, epd, r, u, 45, 7, 20)


def SetMemoryS(dest, modtype, value) -> tuple[Action, Action]:  # noqa: N802
    modtype = EncodeModifier(modtype)
    cpo, unit = rand(dest)
    return (
        SetMemoryC(0x6509B0, Add, cpo),
        SetDeaths(cp.CP, modtype, value, unit),
    )


def MoveCP(dest) -> Action:  # noqa: N802
    try:
        value = dest - _seed
    except (TypeError, AttributeError):
        value = _seed * (-1) + dest
    return SetMemoryC(0x6509B0, Add, value)


def SetMemoryC(dest, modtype, value) -> Action:  # noqa: N802
    modtype = EncodeModifier(modtype)
    _loc = random.randint(0, 0xFFFFFFFF)
    u = random.randint(234, 65535)
    epd = EPD(dest) - 12 * u
    if dest == 0x6509B0:
        global _seed
        if modtype == 7:
            _seed = value
        elif modtype == 9:
            try:
                _seed = _seed - value
            except (TypeError, AttributeError):
                _seed = -value + _seed
        else:
            try:
                _seed = _seed + value
            except (TypeError, AttributeError):
                _seed = value + _seed
    flag = random.randint(0, 0xFF) & (0xFF - 2)
    return Action(_loc, 0, 0, 0, epd, value, u, 45, modtype, flag)



from helper import *


@TestInstance
def test_operator_lvalue():
    a, b = EUDVariable(), EUDVariable()
    with expect_eperror():
        (a + b) << 5

    with expect_eperror():
        (a * b) << 5


@TestInstance
def test_function_lvalue():
    with expect_eperror():
        a = f_dwread_epd(0)
        a << 5


@TestInstance
def test_rvalue():
    from eudplib.core.variable.evcommon import _addor

    v1 = EUDVariable(0)

    trg = NextTrigger()
    trg_count = GetTriggerCounter()
    v2 = v1 + 1 + EUDVariable(2) + 3 + EUDVariable(4) + 5
    ep_assert(trg_count + 5 == GetTriggerCounter(), "Failed to optimize rvalue")
    test_equality(
        "rvalue reuse test: self, __add__",
        [v2, f_dwread(trg + 344)],
        [15, EPD(v2.getValueAddr())],
    )

    trg = NextTrigger()
    v3 = v2 + EUDVariable(6)
    test_equality(
        "rvalue reuse test: other, __add__",
        [v3, f_dwread(trg + 4), f_dwread(trg + 344), f_dwread(trg + 348)],
        [21, v2.GetVTable(), EPD(v2.getDestAddr()), EPD(v3.getValueAddr())],
    )

    trg = NextTrigger()
    trg_count = GetTriggerCounter()
    v4 = EUDVariable(20) - v2
    ep_assert(
        trg_count + 1 == GetTriggerCounter(),
        f"{trg_count} + 1 != {GetTriggerCounter()}",
    )
    nptr = f_dwread(v2.GetVTable() + 4)
    test_equality(
        "rvalue reuse test: self, __sub__",
        [v4, nptr, f_dwread(trg + 344), f_dwread(trg + 348), f_dwread(trg + 4)],
        [
            5,
            _addor.GetVTable(),
            EPD(v4.getValueAddr()),
            1,
            v2.GetVTable(),
        ],
    )

    trg = NextTrigger()
    trg_count = GetTriggerCounter()
    v5 = v3 - EUDVariable(9)
    ep_assert(
        trg_count + 1 == GetTriggerCounter(),
        f"{trg_count} + 1 != {GetTriggerCounter()}",
    )
    test_equality(
        "rvalue reuse test: other, __sub__",
        [
            v5,
            f_dwread(trg + 4),
            f_dwread(trg + 328),
            f_dwread(trg + 344),
            f_dwread(trg + 348),
        ],
        [12, v3.GetVTable(), 0x55555555, EPD(v5.getValueAddr()), 0xFFFFFFFF],
    )

    with expect_eperror():
        SetVariables(f_dwread_epd(0), 0)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from typing import Any

from ..localize import _
from .eperror import EPError, TriggerScopeError, ep_assert

block = tuple[str, Any]


class BlockStruManager:
    def __init__(self) -> None:
        self._blockstru: list[block] = []
        self._lastblockdict: dict[str, list[block]] = {}

    def empty(self) -> bool:
        return not self._blockstru


_current_bsm = BlockStruManager()  # Default one


def set_blockstru_manager(bsm: BlockStruManager) -> BlockStruManager:
    global _current_bsm
    old_bsm = _current_bsm
    _current_bsm = bsm
    return old_bsm


def EUDCreateBlock(name: str, userdata: Any) -> None:  # noqa: N802
    _blockstru = _current_bsm._blockstru
    _lastblockdict = _current_bsm._lastblockdict

    block = (name, userdata)
    _blockstru.append(block)

    if name not in _lastblockdict:
        _lastblockdict[name] = []
    _lastblockdict[name].append(block)


def EUDGetLastBlock() -> block:  # noqa: N802
    _blockstru = _current_bsm._blockstru
    return _blockstru[-1]


def EUDGetLastBlockOfName(name: str) -> block:  # noqa: N802
    _lastblockdict = _current_bsm._lastblockdict
    try:
        return _lastblockdict[name][-1]
    except (IndexError, KeyError):
        if name == "triggerscope":
            raise TriggerScopeError()
        raise EPError(_("Block not found: {}").format(name))


def EUDPeekBlock(name: str) -> block:  # noqa: N802
    lastblock = EUDGetLastBlock()
    ep_assert(
        lastblock[0] == name,
        _("Block starting/ending mismatch")
        + ("\n" + _("    - Started with {}").format(lastblock[0]))
        + ("\n" + _("    - Ended with {}").format(name)),
    )
    return lastblock


def EUDPopBlock(name: str) -> block:  # noqa: N802
    _blockstru = _current_bsm._blockstru
    _lastblockdict = _current_bsm._lastblockdict

    lastblock = _blockstru.pop()
    ep_assert(
        lastblock[0] == name,
        _("Block starting/ending mismatch")
        + ("\n" + _("    - Started with {}").format(lastblock[0]))
        + ("\n" + _("    - Ended with {}").format(name)),
    )
    _lastblockdict[name].pop()
    return lastblock


def EUDGetBlockList() -> list[block]:  # noqa: N802
    return _current_bsm._blockstru



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from typing import Literal, overload

from .. import core as c
from ..core import (
    Action,
    Condition,
    ConstExpr,
    EUDLightBool,
    EUDLightVariable,
    EUDVariable,
    Forward,
)
from ..localize import _
from ..utils import EPD, EPError, ExprProxy, ep_warn, unProxy
from .filler import (
    _filldw,
    _fillhibyte,
    _filllobyte,
    _fillloword,
    _filllsbyte,
    _fillmsbyte,
    _flush_filler,
)


def apply_patch_table(initepd, obj, patch_table: list[list[int | None]]) -> None:
    field_name = 0
    for i, patch_fields in enumerate(patch_table):
        for field_size in patch_fields:
            if isinstance(field_size, int):
                field = obj.fields[field_name]
                if c.IsEUDVariable(field):
                    memory_filler = {
                        -1: _filldw,
                        0: _fillloword,
                        # TODO: add _fillhiword? (ex. set action opcode & modifier)
                        2: _filllsbyte,
                        3: _filllobyte,
                        4: _fillhibyte,
                        5: _fillmsbyte,
                    }[field_size]
                    memory_filler(initepd + i, field)
                    obj.fields[field_name] = 0
            field_name += 1
    _flush_filler()


condpt: list[list[int | None]] = [
    [-1],
    [-1],
    [-1],
    [0, 4, 5],
    [2, 3, None, None],
]

actpt: list[list[int | None]] = [
    [-1],
    [-1],
    [-1],
    [-1],
    [-1],
    [-1],
    [0, 4, 5],
    [2, None, None, None],
]

ConstCondition = (
    bool
    | int
    | EUDVariable
    | EUDLightVariable
    | EUDLightBool
    | ExprProxy[bool | int | EUDVariable | EUDLightVariable | EUDLightBool]
)
_Condition = ConstExpr | ConstCondition | ExprProxy[ConstExpr]


def is_castable(cond) -> bool:
    # EUDArray, EUDVArray, EUDStruct, DBString
    return isinstance(cond, ExprProxy) and type(cond) is not ExprProxy


def patch_condition(cond: _Condition) -> Condition:
    castable = is_castable(cond)
    condition = unProxy(cond)
    if isinstance(condition, Forward):
        if not condition.IsSet():
            raise EPError(_("Forward not initialized"))
        condition = condition.expr
    if isinstance(condition, EUDLightBool):
        return c.MemoryX(condition.getValueAddr(), c.AtLeast, 1, condition._mask)
    if isinstance(condition, (EUDVariable, EUDLightVariable)):  # noqa: UP038
        return c.Memory(condition.getValueAddr(), c.AtLeast, 1)

    # translate boolean condition
    if isinstance(condition, (bool, int)):  # noqa: UP038
        return c.Always() if condition else c.Never()
    if isinstance(condition, Condition):
        apply_patch_table(EPD(condition), condition, condpt)
        return condition
    if castable and isinstance(condition, (ConstExpr, c.RlocInt_C)):  # noqa: UP038
        ep_warn(_("Condition is always True"))
        return c.Always() if condition != 0 else c.Never()
    raise EPError(_("Invalid input for condition: {}").format(cond))


def patch_action(act: Action | Forward | ExprProxy[Action | Forward]) -> Action:
    action = unProxy(act)
    if isinstance(action, Forward):
        if action._expr is None:
            raise EPError(_("Forward not initialized"))
        action = action._expr  # type: ignore[assignment]
    if not isinstance(action, Action):
        raise EPError(_("Action expected, found {}").format(act))
    apply_patch_table(EPD(action), action, actpt)
    return action


@overload
def is_const_cond(cond: Condition | ExprProxy[Condition]) -> bool:
    ...


@overload
def is_const_cond(cond: ConstCondition) -> Literal[True]:
    ...


def is_const_cond(cond) -> bool:
    castable = is_castable(cond)
    cond = unProxy(cond)
    if isinstance(cond, Forward):
        if cond._expr is None:
            return False
        cond = cond._expr
    if isinstance(cond, (bool, int, EUDVariable, EUDLightVariable, EUDLightBool)):  # noqa: UP038
        return True

    if isinstance(cond, Condition):
        field_name = 0
        for cond_fields in condpt:
            for field_size in cond_fields:
                if isinstance(field_size, int):
                    field = cond.fields[field_name]
                    if c.IsEUDVariable(field):
                        return False
                field_name += 1
        return True
    if castable and isinstance(cond, (ConstExpr, c.RlocInt_C)):  # noqa: UP038
        return True
    return False


@overload
def is_nagatable_cond(cond: Condition | ExprProxy[Condition]) -> bool:
    ...


@overload
def is_nagatable_cond(cond: ConstCondition) -> Literal[True]:
    ...


def is_nagatable_cond(cond) -> bool:
    castable = is_castable(cond)
    cond = unProxy(cond)
    if isinstance(cond, Forward):
        if cond._expr is None:
            return False
        cond = cond._expr
    if isinstance(cond, (bool, int, EUDVariable, EUDLightVariable, EUDLightBool)):  # noqa: UP038
        return True

    if isinstance(cond, Condition):
        condtype = cond.fields[5]
        if not isinstance(condtype, int):
            return False
        comparison_set = (1, 2, 3, 4, 5, 12, 14, 15, 21)
        always_or_never = (0, 22, 13, 23)
        if condtype in always_or_never:
            return True
        comparison = cond.fields[4]
        if not isinstance(comparison, int):
            return False
        if condtype == 11 and comparison in (2, 3):  # Switch
            return True
        amount = cond.fields[2]
        if condtype in comparison_set and isinstance(amount, int):
            bring_or_command = (2, 3)
            amount &= 0xFFFFFFFF
            if comparison == 10 and amount == 0:
                return True
            if comparison == 0 and amount <= 1:
                return True
            # AtMost and Exactly/AtLeast behaves differently in Bring/Command
            # (AtMost counts buildings on construction and does not count Egg)
            # So only exchanging (Exactly, 0) <-> (AtLeast, 1) is sound.
            #
            # See: https://cafe.naver.com/edac/book5095361/96809
            if condtype in bring_or_command:
                return False
            if comparison in (0, 1):
                return True
            elif comparison != 10:
                return False
            elif (
                condtype == 15
                and isinstance(cond.fields[8], int)
                and cond.fields[8] == 0x4353
            ):
                mask = cond.fields[0]
                if not isinstance(mask, int):
                    return False
                mask &= 0xFFFFFFFF
                if amount & (~mask):  # never
                    return True
                if amount == mask:
                    return True
                return False
            elif amount == 0xFFFFFFFF:
                return True
        return False
    if castable and isinstance(cond, (ConstExpr, c.RlocInt_C)):  # noqa: UP038
        return True
    return False


def negate_cond(cond: _Condition) -> Condition:
    castable = is_castable(cond)
    condition = unProxy(cond)
    if isinstance(condition, Forward):
        if condition._expr is None:
            raise EPError(_("Forward not initialized"))
        condition = condition._expr
    if isinstance(condition, (EUDVariable, EUDLightVariable)):  # noqa: UP038
        return condition == 0
    if isinstance(condition, EUDLightBool):
        return condition.IsCleared()

    # translate boolean condition
    if isinstance(condition, (bool, int)):  # noqa: UP038
        return c.Never() if condition else c.Always()

    if isinstance(condition, Condition):
        condition.negate()
        return condition
    if castable and isinstance(condition, (ConstExpr, c.RlocInt_C)):  # noqa: UP038
        ep_warn(_("Condition is always False"))
        return c.Never() if condition == 0 else c.Always()
    raise EPError(_("Invalid input for condition: {}").format(cond))



from eudplib import *


def test_static_scmdstr():
    assert SCMD2Text("<04>Test<06>") == "\x04Test\x06"
    assert SCMD2Text("<<04>Test<06>") == "<\x04Test\x06"
    assert SCMD2Text("<04><T>est<06>") == "\x04<T>est\x06"
    assert SCMD2Text("<04>Test<06>") == "\x04Test\x06"
    assert SCMD2Text("<04>Test<06>") == "\x04Test\x06"


test_static_scmdstr()



from helper import *


@TestInstance
def test_stack():
    k = EUDStack()(1024)
    k.push(123)
    k.push(456)
    t1 = k.pop()  # 456
    k.push(789)
    k.push(234)
    t2 = k.pop()  # 234
    t3 = EUDTernary(k.empty())(1)(0)  # 0
    t4 = k.pop()  # 789
    t5 = k.pop()  # 123
    t6 = EUDTernary(k.empty())(1)(0)  # 1

    test_equality("Stack test", [t1, t2, t3, t4, t5, t6], [456, 234, 0, 789, 123, 1])



# Copyright 2023 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from collections.abc import Iterator

def generate_linetable(
    data: bytes, linetable: bytes, positions: Iterator
) -> bytes: ...



# Copyright 2023 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Final, Literal, Self, TypeAlias, overload

from ... import core as c
from ... import utils as ut
from ...core import EUDVariable
from ...core.rawtrigger.consttype import Byte
from ...localize import _
from .epdoffsetmap import EPDOffsetMap
from .memberkind import (
    AnimationKind,
    BaseKind,
    Bit0Kind,
    Bit1Kind,
    ButtonSetKind,
    ByteKind,
    CSpriteKind,
    CUnitKind,
    DamageTypeKind,
    DrawingFunctionKind,
    DwordKind,
    ExplosionTypeKind,
    FlingyKind,
    IconKind,
    ImageKind,
    IscriptKind,
    MapStringKind,
    MovementControlKind,
    PlayerKind,
    PortraitKind,
    PositionKind,
    PositionXKind,
    PositionYKind,
    RaceResearchKind,
    RankKind,
    RightClickActionKind,
    SfxDataKind,
    SpriteKind,
    StatTextKind,
    TechKind,
    UnitKind,
    UnitOrderKind,
    UnitSizeKind,
    UpgradeKind,
    WeaponBehaviorKind,
    WeaponKind,
    WordKind,
    WorkerCarryTypeKind,
)

if TYPE_CHECKING:
    from ...core.rawtrigger.typehint import (
        ButtonSet,
        Rank,
        SfxData,
        String,
        Unit,
        _Flingy,
        _Icon,
        _Image,
        _Iscript,
        _Portrait,
        _Sprite,
        _StatText,
        _Tech,
        _UnitOrder,
        _Upgrade,
        _Weapon,
    )
    from ..csprite import CSprite
    from ..cunit import CUnit
    from ..flingy import Flingy
    from ..image import Image
    from ..player import TrgPlayer
    from ..sprite import Sprite
    from ..tech import Tech
    from ..unit import TrgUnit
    from ..unitorder import UnitOrder
    from ..upgrade import Upgrade
    from ..weapon import Weapon


class BaseMember(metaclass=ABCMeta):
    """Base descriptor class/mixin for EPDOffsetMap"""

    __slots__ = ()
    layout: Final[Literal["struct", "array"]]
    offset: Final[int]
    stride: Final[int]

    def __init__(
        self,
        layout: Literal["struct", "array"],
        offset: int,
        *,
        stride: int | None = None,
    ) -> None:
        self.layout = layout  # type: ignore[misc]
        self.offset = offset  # type: ignore[misc]
        size = self.kind.size()
        ut.ep_assert(offset % 4 + size <= 4, _("Malaligned member"))
        self.stride = size if stride is None else stride  # type: ignore[misc]
        ut.ep_assert(size <= self.stride, _("stride should be at least member size"))

    @property
    @abstractmethod
    def kind(self) -> type[BaseKind]:
        ...

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]):
        ...

    def __get__(self, instance, owner):
        if instance is None:
            return self
        epd, subp = self._get_epd(instance)
        return self.kind.cast(self.kind.read_epd(epd, subp))

    def __set__(self, instance: EPDOffsetMap, value) -> None:
        epd, subp = self._get_epd(instance)
        self.kind.write_epd(epd, subp, self.kind.cast(value))

    def __set_name__(self, owner: type[EPDOffsetMap], name: str) -> None:
        self.__objclass__ = owner  # type: ignore[misc]
        self.__name__ = name  # type: ignore[misc]
        if self.layout == "array":
            owner._has_array_member = True

    def _get_epd(
        self, instance: EPDOffsetMap
    ) -> tuple[int | EUDVariable, int | EUDVariable]:
        if self.layout == "array":
            value = instance._value
            if not isinstance(value, EUDVariable):
                epd, subp = divmod(self.offset + value * self.stride, 4)
                return ut.EPD(0) + epd, subp
            if self.stride == 4:
                return ut.EPD(self.offset) + value, self.offset % 4

            # lazy calculate multiplication & division
            q, r, update = instance._get_stride_cache(self.stride)
            update_start, update_restore, update_end = update

            nexttrg = c.Forward()
            c.RawTrigger(
                nextptr=update_start,
                actions=[
                    c.SetNextPtr(update_start, update_restore),
                    c.SetMemory(update_start + 348, c.SetTo, nexttrg),
                    c.SetNextPtr(update_end, nexttrg),
                ],
            )
            nexttrg << c.NextTrigger()

            epd = ut.EPD(self.offset) + q
            if self.offset % 4 == 0:
                subp = r
            else:
                subp = self.offset % 4 + r
                if self.offset % 4 + (4 - self.stride) % 4 >= 4:
                    c.RawTrigger(
                        conditions=subp.AtLeast(4),
                        actions=[subp.SubtractNumber(4), epd.AddNumber(1)],
                    )
            return epd, subp

        elif self.layout == "struct":
            quotient, remainder = divmod(self.offset, 4)
            return instance._value + quotient, remainder

        raise NotImplementedError


class UnsupportedMember(BaseMember):
    """
    'Sorry, this EUD map is not supported' error is raised when it's accessed.

    '이 EUD 지도는 스타크래프트 리마스터와 호환되지 않습니다.' 오류가 발생합니다.
    """

    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return BaseKind

    def __init__(
        self,
        layout: Literal["struct", "array"],
        offset: int,
        *,
        stride: int | None = None,
    ) -> None:
        self.layout = layout  # type: ignore[misc]
        self.offset = offset  # type: ignore[misc]
        self.stride = 0 if stride is None else stride  # type: ignore[misc]

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]):
        ...

    def __get__(self, instance, owner):
        if instance is None:
            return self
        raise ut.EPError(
            _(
                "StarCraft: Remastered does not support {} "
                "and causes in-game errors."
            ).format(f"{self.__objclass__}.{self.__name__}")
        )

    def __set__(self, instance: EPDOffsetMap, value) -> None:
        raise ut.EPError(
            _(
                "StarCraft: Remastered does not support {} "
                "and causes in-game errors."
            ).format(f"{self.__objclass__}.{self.__name__}")
        )


class NotImplementedMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return BaseKind

    def __init__(
        self,
        layout: Literal["struct", "array"],
        offset: int,
        *,
        stride: int | None = None,
    ) -> None:
        self.layout = layout  # type: ignore[misc]
        self.offset = offset  # type: ignore[misc]
        self.stride = 0 if stride is None else stride  # type: ignore[misc]

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]):
        ...

    def __get__(self, instance, owner):
        if instance is None:
            return self
        raise ut.EPError(
            _("{} is not implemented for {}").format(
                self.__name__, self.__objclass__
            )
        )

    def __set__(self, instance: EPDOffsetMap, value) -> None:
        raise ut.EPError(
            _("{} is not implemented for {}").format(
                self.__name__, self.__objclass__
            )
        )


class BoolMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return Bit0Kind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)


class Bit1Member(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return Bit1Kind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)


class ByteMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return ByteKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)


class PlayerMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return PlayerKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> TrgPlayer:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)


class WeaponMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return WeaponKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]) -> Weapon:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: _Weapon) -> None:
        super().__set__(instance, value)


class UnitOrderMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return UnitOrderKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> UnitOrder:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: _UnitOrder) -> None:
        super().__set__(instance, value)


class UpgradeMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return UpgradeKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]) -> Upgrade:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: _Upgrade) -> None:
        super().__set__(instance, value)


class TechMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return TechKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]) -> Tech:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: _Tech) -> None:
        super().__set__(instance, value)


UnitSize: TypeAlias = Literal["Independent", "Small", "Medium", "Large"] | Byte


class UnitSizeMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return UnitSizeKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: UnitSize) -> None:
        super().__set__(instance, value)


RightClickAction: TypeAlias = (
    Literal[
        "NoCommand_AutoAttack",
        "NormalMove_NormalAttack",
        "NormalMove_NoAttack",
        "NoMove_NormalAttack",
        "Harvest",
        "HarvestAndRepair",
        "Nothing",
    ]
    | Byte
)


class RightClickActionMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return RightClickActionKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: RightClickAction) -> None:
        super().__set__(instance, value)


MovementControl: TypeAlias = (
    Literal["FlingyDat", "PartiallyMobile_Weapon", "IscriptBin"] | Byte
)


class MovementControlMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return MovementControlKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: MovementControl) -> None:
        super().__set__(instance, value)


DamageType: TypeAlias = (
    Literal["Independent", "Explosive", "Concussive", "Normal", "IgnoreArmor"] | Byte
)


class DamageTypeMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return DamageTypeKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: DamageType) -> None:
        super().__set__(instance, value)


WeaponBehavior: TypeAlias = (
    Literal[
        "Fly_DoNotFollowTarget",
        "Fly_FollowTarget",
        "AppearOnTargetUnit",
        "PersistOnTargetSite",
        "AppearOnTargetSite",
        "AppearOnAttacker",
        "AttackAndSelfDestruct",
        "Bounce",
        "AttackNearbyArea",
        "GoToMaxRange",
    ]
    | Byte
)


class WeaponBehaviorMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return WeaponBehaviorKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: WeaponBehavior) -> None:
        super().__set__(instance, value)


ExplosionType: TypeAlias = (
    Literal[
        "None",
        "NormalHit",
        "SplashRadial",
        "SplashEnemy",
        "Lockdown",
        "NuclearMissile",
        "Parasite",
        "Broodlings",
        "EmpShockwave",
        "Irradiate",
        "Ensnare",
        "Plague",
        "StasisField",
        "DarkSwarm",
        "Consume",
        "YamatoGun",
        "Restoration",
        "DisruptionWeb",
        "CorrosiveAcid",
        "MindControl",
        "Feedback",
        "OpticalFlare",
        "Maelstrom",
        "Unknown_Crash",
        "SplashAir",
    ]
    | Byte
)


class ExplosionTypeMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return ExplosionTypeKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: ExplosionType) -> None:
        super().__set__(instance, value)


DrawingFunction: TypeAlias = (
    Literal[
        "Normal",
        "NormalNoHallucination",
        "NonVisionCloaking",
        "NonVisionCloaked",
        "NonVisionDecloaking",
        "VisionCloaking",
        "VisionCloaked",
        "VisionDecloaking",
        "EMPShockwave",
        "UseRemapping",
        "Shadow",
        "HpBar",
        "WarpTexture",
        "SelectionCircle",
        "PlayerColorOverride",
        "HideGFX_ShowSizeRect",
        "Hallucination",
        "WarpFlash",
    ]
    | Byte
)


class DrawingFunctionMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return DrawingFunctionKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: DrawingFunction) -> None:
        super().__set__(instance, value)


IscriptAnimation: TypeAlias = (
    Literal[
        "Init",
        "Death",
        "GndAttkInit",
        "AirAttkInit",
        "Unused1",
        "GndAttkRpt",
        "AirAttkRpt",
        "CastSpell",
        "GndAttkToIdle",
        "AirAttkToIdle",
        "Unused2",
        "Walking",
        "WalkingToIdle",
        "SpecialState1",
        "SpecialState2",
        "AlmostBuilt",
        "Built",
        "Landing",
        "LiftOff",
        "IsWorking",
        "WorkingToIdle",
        "WarpIn",
        "Unused3",
        "StarEditInit",
        "Disable",
        "Burrow",
        "UnBurrow",
        "Enable",
        "NoAnimation",
    ]
    | Byte
)


class AnimationMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return AnimationKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: IscriptAnimation) -> None:
        super().__set__(instance, value)


RaceResearch: TypeAlias = Literal["Zerg", "Terran", "Protoss", "All"] | Byte


class RaceResearchMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return RaceResearchKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: RaceResearch) -> None:
        super().__set__(instance, value)


WorkerCarryType: TypeAlias = (
    Literal["None", "Gas", "Ore", "GasOrOre", "PowerUp"] | Byte
)


class WorkerCarryTypeMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return WorkerCarryTypeKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: WorkerCarryType) -> None:
        super().__set__(instance, value)


class RankMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return RankKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: Rank) -> None:
        super().__set__(instance, value)


class FlingyMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return FlingyKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]) -> Flingy:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: _Flingy) -> None:
        super().__set__(instance, value)


class WordMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return WordKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)


class PositionXMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return PositionXKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)


class PositionYMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return PositionYKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)


class UnitMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return UnitKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]) -> TrgUnit:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: Unit) -> None:
        super().__set__(instance, value)


class SpriteMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return SpriteKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]) -> Sprite:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: _Sprite) -> None:
        super().__set__(instance, value)


class ImageMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return ImageKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]) -> Image:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: _Image) -> None:
        super().__set__(instance, value)


class StatTextMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return StatTextKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: _StatText) -> None:
        super().__set__(instance, value)


class SfxDataMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return SfxDataKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: SfxData) -> None:
        super().__set__(instance, value)


class PortraitMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return PortraitKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: _Portrait) -> None:
        super().__set__(instance, value)


class IconMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return IconKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: _Icon) -> None:
        super().__set__(instance, value)


class MapStringMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return MapStringKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: String) -> None:
        super().__set__(instance, value)


class ButtonSetMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return ButtonSetKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: ButtonSet) -> None:
        super().__set__(instance, value)


class DwordMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return DwordKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)


class PositionMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return PositionKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)


class CUnitMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return CUnitKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]) -> CUnit:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)


class CSpriteMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return CSpriteKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]) -> CSprite:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)


class IscriptMember(BaseMember):
    __slots__ = ("layout", "offset", "stride", "__objclass__", "__name__")

    @property
    def kind(self):
        return IscriptKind

    @overload
    def __get__(self, instance: None, owner: type[EPDOffsetMap]) -> Self:
        ...

    @overload
    def __get__(
        self, instance: EPDOffsetMap, owner: type[EPDOffsetMap]
    ) -> EUDVariable:
        ...

    def __get__(self, instance, owner):
        return super().__get__(instance, owner)

    def __set__(self, instance: EPDOffsetMap, value: _Iscript) -> None:
        super().__set__(instance, value)



import random

from helper import *


@TestInstance
def test_dwmemio():
    a = Db(i2b4(0x1234))

    # Normal reading
    ptr, epd = f_dwepdread_epd(EPD(a))

    # For any ptr, there are 4 valid epd values. So comparing epd and epd
    # for equality won't work. Here we convert epd back to ptr for comparison.
    test_equality("f_dwread_epd", [ptr, 0x58A364 + 4 * epd], [0x1234, 0x1234])

    # Map position reading
    b1 = random.randint(0, 0xFFFFFFFF)
    dim = GetChkTokenized().getsection("DIM")
    bx, by = b2i2(dim), b2i2(dim, 2)
    bx = 2 ** ((bx - 1).bit_length() + 5) - 1
    by = 2 ** ((by - 1).bit_length() + 5) - 1
    bx, by = b1 & bx, (b1 & (by << 16)) >> 16

    x, y = f_posread_epd(EPD(Db(i2b4(b1))))
    test_equality("f_posread_epd %X" % b1, [x, y], [bx, by])

    # CUnit reading
    c1 = random.randint(0, 0xFFFFFFFF)
    if c1 == 0:
        cptr, cepd = 0, 0
    else:
        cptr, cepd = 0x400008, 0xFFF9D729
        for bit in range(4, 22):
            bit = 1 << bit
            if c1 & bit:
                cptr += bit
                cepd += bit >> 2
        cepd = cepd & 0xFFFFFFFF

    ptr, epd = f_cunitepdread_epd(EPD(Db(i2b4(c1))))
    test_equality("f_cunitread_epd %X" % c1, [ptr, epd], [cptr, cepd])
    ptr, epd = f_cunitepdread_epd(EPD(Db(4)))
    test_equality("f_cunitread_epd -> 0 values", [ptr, epd], (0, 0))
    oldcp = f_getcurpl()
    f_dwwrite(0x6509B0, EPD(Db(i2b4(c1))))
    ptr, epd = f_cunitepdread_cp(0)
    f_setcurpl(oldcp)
    test_equality("f_cunitread_cp %X" % c1, [ptr, epd], [cptr, cepd])
    f_dwwrite(0x6509B0, EPD(Db(4)))
    ptr, epd = f_cunitepdread_cp(0)
    f_setcurpl(oldcp)
    test_equality("f_cunitread_cp -> 0 values", [ptr, epd], (0, 0))

    # Flag reading!
    f1, f2, f3, f4 = f_flagread_epd(EPD(a), 0x1000, 0x0100, ~0x0010, 0x0004)
    test_equality(
        "f_flagread_epd", [f1, f2, f3, f4], [0x1000, 0x0000, 0x1224, 0x0004]
    )

    # dwwrite
    f_dwwrite_epd(EPD(a), 1234)
    test_assert("f_dwwrite works", Memory(a, Exactly, 1234))

    # Fixes euddraft/issues#4
    f_setcurpl(EPD(a))
    f_dwsubtract_cp(0, 12)
    f_setcurpl(Player1)
    test_assert("f_dwwrite works", Memory(a, Exactly, 1222))

    f_srand(0)
    test_equality("f_rand works", [f_rand(), f_getseed()], [0, 12345])
    test_equality("f_dwrand works", [f_dwrand(), f_getseed()], [0xD3DCA704, 0xA70427DF])

    EP_SetRValueStrictMode(False)
    f_randomize()
    EP_SetRValueStrictMode(True)

    # Fixes https://cafe.naver.com/edac/81090
    SetKills(CurrentPlayer, SetTo, 0, 0)



# Copyright 2023 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

# ruff: noqa: N815
from typing import ClassVar, TypeAlias, TypeVar, cast

from .. import core as c
from ..localize import _
from ..utils import EPD, EPError, ExprProxy, classproperty, unProxy
from .offsetmap import (
    ByteEnumMember,
    ByteMember,
    CSpriteMember,
    DwordMember,
    EPDOffsetMap,
    Flag,
    PlayerMember,
    PositionMember,
    PositionXMember,
    PositionYMember,
    SpriteMember,
    WordMember,
)
from .offsetmap.epdoffsetmap import _epd_cache, _ptr_cache


class CSpriteFlags(ByteEnumMember):
    __slots__ = ()
    DrawSelCircle = Flag(0x01)
    "Draw selection circle"
    AllySel1 = Flag(0x02)
    AllySel2 = Flag(0x04)
    Selected = Flag(0x08)
    "Draw HP bar, Selected"
    IsSubunit = Flag(0x10)
    "sorts sprite elevation higher, so that subunits always appear above base unit"
    Hidden = Flag(0x20)
    Burrowed = Flag(0x40)
    IscriptCode = Flag(0x80)
    "Iscript unbreakable code section"


T = TypeVar("T", bound="CSprite")
int_or_var: TypeAlias = int | c.EUDVariable | ExprProxy


class CSprite(EPDOffsetMap):
    __slots__ = "_ptr"
    prev: ClassVar[CSpriteMember] = CSpriteMember("struct", 0x00)
    next: ClassVar[CSpriteMember] = CSpriteMember("struct", 0x04)
    sprite: ClassVar[SpriteMember] = SpriteMember("struct", 0x08)
    player: ClassVar[PlayerMember] = PlayerMember("struct", 0x0A)
    "officially 'creator'"
    selectionIndex: ClassVar[ByteMember] = ByteMember("struct", 0x0B)
    "0 <= selectionIndex <= 11. Index in the selection area at bottom of screen."
    visibilityFlags: ClassVar[ByteMember] = ByteMember("struct", 0x0C)
    "Player bits indicating visibility for a player (not hidden by the fog-of-war)"
    elevationLevel: ClassVar[ByteMember] = ByteMember("struct", 0x0D)
    flags = CSpriteFlags("struct", 0x0E)
    selectionTimer: ClassVar[ByteMember] = ByteMember("struct", 0x0F)
    index: ClassVar[WordMember] = WordMember("struct", 0x10)
    grpWidth: ClassVar[ByteMember] = ByteMember("struct", 0x12)
    grpHeight: ClassVar[ByteMember] = ByteMember("struct", 0x13)
    pos: ClassVar[PositionMember] = PositionMember("struct", 0x14)
    posX: ClassVar[PositionXMember] = PositionXMember("struct", 0x14)
    posY: ClassVar[PositionYMember] = PositionYMember("struct", 0x16)
    mainGraphic: ClassVar[DwordMember] = DwordMember("struct", 0x18)
    "officially 'pImagePrimary', CImage"
    imageHead: ClassVar[DwordMember] = DwordMember("struct", 0x1C)
    imageTail: ClassVar[DwordMember] = DwordMember("struct", 0x20)

    @classproperty
    def range(self):
        return (EPD(0x629D98), EPD(0x629D98) + 9 * 2499, 9)

    def __init__(self, epd: int_or_var, *, ptr: int_or_var | None = None) -> None:
        """EPD Constructor of CSprite. Use CSprite.from_ptr(ptr) for ptr value"""
        _epd: int | c.EUDVariable
        self._ptr: int | c.EUDVariable | None

        if not isinstance(epd, CSprite):
            u, p = unProxy(epd), unProxy(ptr)
        else:
            u, p = epd._value, epd._ptr

        if isinstance(u, int):
            q, r = divmod(u - EPD(0x629D98), 9)  # check epd
            if r == 0 and 0 <= q < 2500:
                _epd, self._ptr = u, 0x629D98 + 36 * q
            else:
                raise EPError(_("Invalid input for CSprite: {}").format(epd))

            if p is not None and not isinstance(p, int) or p != self._ptr:
                raise EPError(_("Invalid input for CSprite.ptr: {}").format(ptr))
        elif isinstance(u, c.EUDVariable):
            if p is None:
                self._ptr = None
                if EPDOffsetMap._cast:
                    _epd = u
                else:
                    _epd = c.EUDVariable()
                    _epd << u

            else:
                if not isinstance(p, c.EUDVariable):
                    raise EPError(_("Invalid input for CSprite.ptr: {}").format(ptr))
                if EPDOffsetMap._cast:
                    _epd, self._ptr = u, p
                else:
                    _epd, self._ptr = c.EUDCreateVariables(2)
                    c.SetVariables((_epd, self._ptr), (u, p))
        else:
            raise EPError(_("Invalid input for CSprite: {}").format(epd))

        EPDOffsetMap._cast = False
        super().__init__(_epd)

    @classmethod
    def from_ptr(cls: type[T], ptr: int_or_var) -> T:
        epd: int | c.EUDVariable
        u = unProxy(ptr)
        # check ptr
        if isinstance(u, int):
            q, r = divmod(u - 0x629D98, 36)
            if r == 0 and 0 <= q < 2500:
                epd = EPD(u)
            else:
                raise EPError(_("Invalid input for CSprite: {}").format(ptr))
        elif isinstance(u, c.EUDVariable):
            epd = _epd_cache(u)
        else:
            raise EPError(_("Invalid input for CSprite: {}").format(epd))

        return cls(epd, ptr=u)

    @classmethod
    def from_read(cls: type[T], epd) -> T:
        from ..memio import f_spriteepdread_epd

        _ptr, _epd = f_spriteepdread_epd(epd)
        return cls(_epd, ptr=_ptr)

    @property
    def ptr(self) -> int | c.EUDVariable:
        if isinstance(self._value, int):
            return cast(int, self._ptr)  # FIXME
        if self._ptr is None:
            self._ptr = c.EUDVariable()
        return _ptr_cache(self._value, self._ptr)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.
from .qgc import (
    QueueGameCommand,
    QueueGameCommand_AddSelect,
    QueueGameCommand_MergeArchon,
    QueueGameCommand_MergeDarkArchon,
    QueueGameCommand_MinimapPing,
    QueueGameCommand_PauseGame,
    QueueGameCommand_QueuedRightClick,
    QueueGameCommand_RemoveSelect,
    QueueGameCommand_RestartGame,
    QueueGameCommand_ResumeGame,
    QueueGameCommand_RightClick,
    QueueGameCommand_Select,
    QueueGameCommand_TrainUnit,
    QueueGameCommand_UseCheat,
)

__all__ = [
    "QueueGameCommand",
    "QueueGameCommand_AddSelect",
    "QueueGameCommand_MergeArchon",
    "QueueGameCommand_MergeDarkArchon",
    "QueueGameCommand_MinimapPing",
    "QueueGameCommand_PauseGame",
    "QueueGameCommand_QueuedRightClick",
    "QueueGameCommand_RemoveSelect",
    "QueueGameCommand_RestartGame",
    "QueueGameCommand_ResumeGame",
    "QueueGameCommand_RightClick",
    "QueueGameCommand_Select",
    "QueueGameCommand_TrainUnit",
    "QueueGameCommand_UseCheat",
]



# Copyright 2019 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ...utils import EPD
from ..allocator import Forward
from ..rawtrigger import Add, SetTo
from .eudv import EUDCreateVariables
from .eudxv import EUDXVariable

_ev = EUDCreateVariables(5)
_xv = [EUDXVariable(0, SetTo, 0, 0) for _ in range(1)]
_addor = EUDXVariable(0, Add, 0)
_selfadder_valueaddr = Forward()
_selfadder = EUDXVariable(EPD(_selfadder_valueaddr), Add, 0)
_selfadder_valueaddr << _selfadder.getValueAddr()



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ... import core as c
from ... import ctrlstru as cs
from ... import utils as ut
from ...core.allocator.pbuffer import Payload
from ...memio import f_dwread_epd

""" Stage 2 :
- Initialize payload (stage3+ + user code) & execute it
"""


def create_payload_relocator(payload: Payload) -> Payload:
    # We first build code injector.
    prtdb = c.Db(b"".join([ut.i2b4(x // 4) for x in payload.prttable]))
    ortdb = c.Db(b"".join([ut.i2b4(x // 4) for x in payload.orttable]))
    n = c.EUDVariable()

    orig_payload = c.Db(payload.data)

    if c.PushTriggerScope():
        root = c.NextTrigger()

        # Verify payload
        # Note : this is very, very basic protection method. Intended attackers
        # should be able to penetrate through this very easily

        # init prt
        if payload.prttable:
            n << len(payload.prttable)
            if cs.EUDWhile()(n >= 1):
                n += ut.EPD(prtdb) - 1
                epd, dst = f_dwread_epd(n), c.Forward()
                c.VProc(
                    epd,
                    [
                        n.SubtractNumber(ut.EPD(prtdb)),
                        epd.AddNumber(ut.EPD(orig_payload)),
                        epd.SetDest(ut.EPD(dst) + 4),
                    ],
                )
                cs.DoActions(dst << c.SetDeaths(0, c.Add, orig_payload // 4, 0))
            cs.EUDEndWhile()

        # init ort
        if payload.orttable:
            n << len(payload.orttable)
            if cs.EUDWhile()(n >= 1):
                n += ut.EPD(ortdb) - 1
                epd, dst = f_dwread_epd(n), c.Forward()
                c.VProc(
                    epd,
                    [
                        n.SubtractNumber(ut.EPD(ortdb)),
                        epd.AddNumber(ut.EPD(orig_payload)),
                        epd.SetDest(ut.EPD(dst) + 4),
                    ],
                )
                cs.DoActions(dst << c.SetDeaths(0, c.Add, orig_payload, 0))
            cs.EUDEndWhile()

        # Jump
        c.SetNextTrigger(orig_payload)

    c.PopTriggerScope()

    return c.CreatePayload(root)



# Copyright 2023 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .enummember import (
    ByteEnumMember,
    DwordEnumMember,
    Flag,
    MovementFlags,
    WordEnumMember,
)
from .epdoffsetmap import EPDOffsetMap
from .member import (
    AnimationMember,
    BaseMember,
    Bit1Member,
    BoolMember,
    ButtonSetMember,
    ByteMember,
    CSpriteMember,
    CUnitMember,
    DamageTypeMember,
    DrawingFunctionMember,
    DwordMember,
    ExplosionTypeMember,
    FlingyMember,
    IconMember,
    ImageMember,
    IscriptMember,
    MapStringMember,
    MovementControlMember,
    NotImplementedMember,
    PlayerMember,
    PortraitMember,
    PositionMember,
    PositionXMember,
    PositionYMember,
    RaceResearchMember,
    RankMember,
    RightClickActionMember,
    SfxDataMember,
    SpriteMember,
    StatTextMember,
    TechMember,
    UnitMember,
    UnitOrderMember,
    UnitSizeMember,
    UnsupportedMember,
    UpgradeMember,
    WeaponBehaviorMember,
    WeaponMember,
    WordMember,
    WorkerCarryTypeMember,
)

__all__ = [
    "ByteEnumMember",
    "DwordEnumMember",
    "Flag",
    "MovementFlags",
    "WordEnumMember",
    "EPDOffsetMap",
    "AnimationMember",
    "BaseMember",
    "Bit1Member",
    "BoolMember",
    "ButtonSetMember",
    "ByteMember",
    "CSpriteMember",
    "CUnitMember",
    "DamageTypeMember",
    "DrawingFunctionMember",
    "DwordMember",
    "ExplosionTypeMember",
    "FlingyMember",
    "IconMember",
    "ImageMember",
    "IscriptMember",
    "MapStringMember",
    "MovementControlMember",
    "NotImplementedMember",
    "PlayerMember",
    "PortraitMember",
    "PositionMember",
    "PositionXMember",
    "PositionYMember",
    "RaceResearchMember",
    "RankMember",
    "RightClickActionMember",
    "SfxDataMember",
    "SpriteMember",
    "StatTextMember",
    "TechMember",
    "UnitMember",
    "UnitOrderMember",
    "UnitSizeMember",
    "UnsupportedMember",
    "UpgradeMember",
    "WeaponBehaviorMember",
    "WeaponMember",
    "WordMember",
    "WorkerCarryTypeMember",
]



from helper import *


@TestInstance
def test_ptrigger():
    if EUDPlayerLoop()():
        PTrigger(
            players=[Player1],
            actions=[
                SetDeaths(CurrentPlayer, Add, 1, "Terran Marine"),
                PreserveTrigger(),
            ],
        )

        PTrigger(
            players=[Player1, Player7],
            actions=[
                SetDeaths(CurrentPlayer, Add, 1, "Terran Marine"),
                PreserveTrigger(),
            ],
        )

        PTrigger(
            players=[Force1],
            actions=[
                SetDeaths(CurrentPlayer, Add, 1, "Terran Marine"),
                PreserveTrigger(),
            ],
        )
    EUDEndPlayerLoop()

    test_equality(
        "PTrigger test",
        [f_dwread_epd(0), f_dwread_epd(6), f_dwread_epd(7)],
        [3, 1, 0],
    )

    DoActions(SetDeaths(AllPlayers, SetTo, 0, "Terran Marine"))



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from collections.abc import Iterator
from typing import Any, Literal

from .. import core as c
from .. import utils as ut
from ..localize import _
from .basicstru import EUDJump, EUDJumpIf, EUDJumpIfNot
from .cshelper import CtrlStruOpener


def _is_loopblock(block: dict[str, Any]) -> bool:
    return "contpoint" in block


# -------


def EUDInfLoop() -> CtrlStruOpener:  # noqa: N802
    def _footer() -> Literal[True]:
        block = {
            "loopstart": c.NextTrigger(),
            "loopend": c.Forward(),
            "contpoint": c.Forward(),
        }

        ut.EUDCreateBlock("infloopblock", block)
        return True

    return CtrlStruOpener(_footer)


def EUDEndInfLoop() -> None:  # noqa: N802
    block = ut.EUDPopBlock("infloopblock")[1]
    if not block["contpoint"].IsSet():
        block["contpoint"] << block["loopstart"]
    c.SetNextTrigger(block["loopstart"])
    block["loopend"] << c.NextTrigger()


# -------


def EUDLoopN() -> CtrlStruOpener:  # noqa: N802
    def _footer(n: int) -> Literal[True]:
        vardb = c.Db(4)
        c.RawTrigger(actions=c.SetMemory(vardb, c.SetTo, n))

        block = {
            "loopstart": c.NextTrigger(),
            "loopend": c.Forward(),
            "contpoint": c.Forward(),
            "vardb": vardb,
        }

        ut.EUDCreateBlock("loopnblock", block)
        EUDJumpIf(c.Memory(vardb, c.Exactly, 0), block["loopend"])
        return True

    return CtrlStruOpener(_footer)


def EUDEndLoopN() -> None:  # noqa: N802
    block = ut.EUDPopBlock("loopnblock")[1]
    if not block["contpoint"].IsSet():
        block["contpoint"] << c.NextTrigger()

    vardb = block["vardb"]
    c.RawTrigger(
        nextptr=block["loopstart"], actions=c.SetMemory(vardb, c.Subtract, 1)
    )
    block["loopend"] << c.NextTrigger()


# -------


def EUDLoopRange(start, end=None) -> Iterator[c.EUDVariable]:  # noqa: N802
    ut.EUDCreateBlock("looprangeblock", None)
    if end is None:
        start, end = 0, start

    v = c.EUDVariable()
    v << start
    if EUDWhileNot()(v >= end):
        block = ut.EUDPeekBlock("whileblock")[1]
        yield v
        if not block["contpoint"].IsSet():
            set_continuepoint()
        v += 1
    EUDEndWhile()
    ut.EUDPopBlock("looprangeblock")


# -------


def EUDWhile() -> CtrlStruOpener:  # noqa: N802
    def _header() -> None:
        block = {
            "loopstart": c.NextTrigger(),
            "loopend": c.Forward(),
            "contpoint": c.Forward(),
            "conditional": True,
        }

        ut.EUDCreateBlock("whileblock", block)

    def _footer(conditions, *, neg=False) -> Literal[True]:
        block = ut.EUDPeekBlock("whileblock")[1]
        if neg:
            EUDJumpIf(conditions, block["loopend"])
        else:
            EUDJumpIfNot(conditions, block["loopend"])
        block["conditional"] = False
        return True

    _header()
    return CtrlStruOpener(_footer)


def EUDWhileNot() -> CtrlStruOpener:  # noqa: N802
    c = EUDWhile()
    return CtrlStruOpener(lambda conditions: c(conditions, neg=True))


def _unsafe_whilenot() -> CtrlStruOpener:
    def _header() -> None:
        block = {
            "loopstart": c.NextTrigger(),
            "loopend": c.Forward(),
            "contpoint": c.Forward(),
            "conditional": True,
        }

        ut.EUDCreateBlock("whileblock", block)

    def _footer(conditions) -> Literal[True]:
        block = ut.EUDPeekBlock("whileblock")[1]
        c.RawTrigger(
            conditions=conditions,
            actions=c.SetNextPtr(block["loopstart"], block["loopend"]),
        )
        block["conditional"] = False
        return True

    _header()
    return CtrlStruOpener(_footer)


def EUDEndWhile() -> None:  # noqa: N802
    block = ut.EUDPopBlock("whileblock")[1]
    if not block["contpoint"].IsSet():
        block["contpoint"] << block["loopstart"]
    c.SetNextTrigger(block["loopstart"])
    block["loopend"] << c.NextTrigger()


# -------


def _get_last_loopblock() -> tuple[str, Any]:
    for block in reversed(ut.EUDGetBlockList()):
        if _is_loopblock(block[1]):
            return block

    raise ut.EPError(_("No loop block surrounding this code area"))


def eudloop_continue() -> None:
    block = _get_last_loopblock()[1]
    try:
        EUDJump(block["contpoint"])
    except ut.EPError:
        ut.ep_warn(_("unreachable continue"))


def eudloop_continue_if(conditions) -> None:
    block = _get_last_loopblock()[1]
    EUDJumpIf(conditions, block["contpoint"])


def eudloop_continue_ifnot(conditions) -> None:
    block = _get_last_loopblock()[1]
    EUDJumpIfNot(conditions, block["contpoint"])


def is_continuepoint_set() -> bool:
    block = _get_last_loopblock()[1]
    return block["contpoint"].IsSet()


def set_continuepoint() -> None:
    block = _get_last_loopblock()[1]
    block["contpoint"] << c.NextTrigger()


def eudloop_break() -> None:
    block = _get_last_loopblock()[1]
    try:
        EUDJump(block["loopend"])
    except ut.EPError:
        ut.ep_warn(_("unreachable break"))


def eudloop_break_if(conditions) -> None:
    block = _get_last_loopblock()[1]
    EUDJumpIf(conditions, block["loopend"])


def eudloop_break_ifnot(conditions) -> None:
    block = _get_last_loopblock()[1]
    EUDJumpIfNot(conditions, block["loopend"])



from helper import *

a = EUDVariable(0)
b = EUDVariable(0)
i = EUDVariable(0)


@EUDFunc
def f_func_with_side_effect(b):
    global a
    a += 1
    return b


@TestInstance
def test_once():
    global a, b, i
    a << 0
    i << 0
    # repeat once block multiple times
    if EUDWhile()(i <= 9):
        if EUDExecuteOnce()():
            f_func_with_side_effect(0)
        EUDEndExecuteOnce()
        EUDSetContinuePoint()
        i += 1
    EUDEndWhile()
    test_equality("EUDExecuteOnce w. (true)", [a, i], [1, 10])

    a << 0
    b << 0
    i << 0
    if EUDWhile()(i <= 9):
        if EUDExecuteOnce()(f_func_with_side_effect(1)):
            b += 1
        EUDEndExecuteOnce()
        EUDSetContinuePoint()
        i += 1
    EUDEndWhile()
    test_equality("EUDExecuteOnce w. (true (side_effect)", [a, b], [1, 1])

    a << 0
    b << 0
    i << 0
    if EUDWhile()(i <= 9):
        if EUDExecuteOnce()(f_func_with_side_effect(0), neg=True):
            b += 1
        EUDEndExecuteOnce()
        EUDSetContinuePoint()
        i += 1
    EUDEndWhile()
    test_equality("EUDExecuteOnce w. (false (side_effect)", [a, b], [1, 1])



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import functools
import inspect
from collections.abc import Callable

from ... import utils as ut
from ...localize import _
from .. import variable as ev
from ..eudstruct.eudstruct import EUDStruct
from ..eudstruct.selftype import _set_selftype, selftype
from .eudtypedfuncn import EUDTypedFuncN, _apply_types

_mth_classtype: dict[Callable, type] = {}


def EUDTypedMethod(argtypes, rettypes=None, *, traced=False):  # noqa: N802
    def _eud_typed_method(method):
        # Get argument number of fdecl_func
        argspec = inspect.getfullargspec(method)
        ut.ep_assert(
            argspec[1] is None,
            _("No variadic arguments (*args) allowed for EUDFunc."),
        )
        ut.ep_assert(
            argspec[2] is None,
            _("No variadic keyword arguments (**kwargs) allowed for EUDFunc."),
        )

        # Get number of arguments excluding self
        argn = len(argspec[0]) - 1

        constexpr_callmap = {}

        # Generic caller
        def generic_caller(self, *args):
            _set_selftype(_mth_classtype[method])
            self = selftype.cast(self)
            args = _apply_types(argtypes, args)
            _set_selftype(None)
            return method(self, *args)

        generic_caller = EUDTypedFuncN(
            argn + 1, generic_caller, method, argtypes, rettypes, traced=traced
        )

        # Return function
        def call(self, *args):
            # Use purely eudfun method
            if isinstance(self, EUDStruct) or ev.IsEUDVariable(self):
                selftype = type(self)
                if method not in _mth_classtype:
                    _mth_classtype[method] = selftype

                _set_selftype(selftype)
                rets = generic_caller(self, *args)  # FIXME: euddraft#34
                _set_selftype(None)
                return rets

            # Const expression. Can use optimizations
            else:
                if self not in constexpr_callmap:

                    def caller(*args):
                        args = _apply_types(argtypes, args)
                        return method(self, *args)

                    constexpr_callmap[self] = EUDTypedFuncN(
                        argn, caller, method, argtypes, rettypes, traced=traced
                    )

                _set_selftype(type(self))
                rets = constexpr_callmap[self](*args)
                _set_selftype(None)
                return rets

        functools.update_wrapper(call, method)
        return call

    return _eud_typed_method


def EUDTracedTypedMethod(argtypes, rettypes=None):  # noqa: N802
    return EUDTypedMethod(argtypes, rettypes, traced=True)


def EUDMethod(method):  # noqa: N802
    return EUDTypedMethod(None, None, traced=False)(method)


def EUDTracedMethod(method):  # noqa: N802
    return EUDTypedMethod(None, None, traced=True)(method)



from __future__ import annotations

import pytest

from pybind11_tests import ConstructorStats
from pybind11_tests import operators as m


def test_operator_overloading():
    v1 = m.Vector2(1, 2)
    v2 = m.Vector(3, -1)
    v3 = m.Vector2(1, 2)  # Same value as v1, but different instance.
    assert v1 is not v3

    assert str(v1) == "[1.000000, 2.000000]"
    assert str(v2) == "[3.000000, -1.000000]"

    assert str(-v2) == "[-3.000000, 1.000000]"

    assert str(v1 + v2) == "[4.000000, 1.000000]"
    assert str(v1 - v2) == "[-2.000000, 3.000000]"
    assert str(v1 - 8) == "[-7.000000, -6.000000]"
    assert str(v1 + 8) == "[9.000000, 10.000000]"
    assert str(v1 * 8) == "[8.000000, 16.000000]"
    assert str(v1 / 8) == "[0.125000, 0.250000]"
    assert str(8 - v1) == "[7.000000, 6.000000]"
    assert str(8 + v1) == "[9.000000, 10.000000]"
    assert str(8 * v1) == "[8.000000, 16.000000]"
    assert str(8 / v1) == "[8.000000, 4.000000]"
    assert str(v1 * v2) == "[3.000000, -2.000000]"
    assert str(v2 / v1) == "[3.000000, -0.500000]"

    assert v1 == v3
    assert v1 != v2
    assert hash(v1) == 4
    # TODO(eric.cousineau): Make this work.
    # assert abs(v1) == "abs(Vector2)"

    v1 += 2 * v2
    assert str(v1) == "[7.000000, 0.000000]"
    v1 -= v2
    assert str(v1) == "[4.000000, 1.000000]"
    v1 *= 2
    assert str(v1) == "[8.000000, 2.000000]"
    v1 /= 16
    assert str(v1) == "[0.500000, 0.125000]"
    v1 *= v2
    assert str(v1) == "[1.500000, -0.125000]"
    v2 /= v1
    assert str(v2) == "[2.000000, 8.000000]"

    cstats = ConstructorStats.get(m.Vector2)
    assert cstats.alive() == 3
    del v1
    assert cstats.alive() == 2
    del v2
    assert cstats.alive() == 1
    del v3
    assert cstats.alive() == 0
    assert cstats.values() == [
        "[1.000000, 2.000000]",
        "[3.000000, -1.000000]",
        "[1.000000, 2.000000]",
        "[-3.000000, 1.000000]",
        "[4.000000, 1.000000]",
        "[-2.000000, 3.000000]",
        "[-7.000000, -6.000000]",
        "[9.000000, 10.000000]",
        "[8.000000, 16.000000]",
        "[0.125000, 0.250000]",
        "[7.000000, 6.000000]",
        "[9.000000, 10.000000]",
        "[8.000000, 16.000000]",
        "[8.000000, 4.000000]",
        "[3.000000, -2.000000]",
        "[3.000000, -0.500000]",
        "[6.000000, -2.000000]",
    ]
    assert cstats.default_constructions == 0
    assert cstats.copy_constructions == 0
    assert cstats.move_constructions >= 10
    assert cstats.copy_assignments == 0
    assert cstats.move_assignments == 0


def test_operators_notimplemented():
    """#393: need to return NotSupported to ensure correct arithmetic operator behavior"""

    c1, c2 = m.C1(), m.C2()
    assert c1 + c1 == 11
    assert c2 + c2 == 22
    assert c2 + c1 == 21
    assert c1 + c2 == 12


def test_nested():
    """#328: first member in a class can't be used in operators"""

    a = m.NestA()
    b = m.NestB()
    c = m.NestC()

    a += 10
    assert m.get_NestA(a) == 13
    b.a += 100
    assert m.get_NestA(b.a) == 103
    c.b.a += 1000
    assert m.get_NestA(c.b.a) == 1003
    b -= 1
    assert m.get_NestB(b) == 3
    c.b -= 3
    assert m.get_NestB(c.b) == 1
    c *= 7
    assert m.get_NestC(c) == 35

    abase = a.as_base()
    assert abase.value == -2
    a.as_base().value += 44
    assert abase.value == 42
    assert c.b.a.as_base().value == -2
    c.b.a.as_base().value += 44
    assert c.b.a.as_base().value == 42

    del c
    pytest.gc_collect()
    del a  # Shouldn't delete while abase is still alive
    pytest.gc_collect()

    assert abase.value == 42
    del abase, b
    pytest.gc_collect()


def test_overriding_eq_reset_hash():
    assert m.Comparable(15) is not m.Comparable(15)
    assert m.Comparable(15) == m.Comparable(15)

    with pytest.raises(TypeError) as excinfo:
        hash(m.Comparable(15))
    assert str(excinfo.value).startswith("unhashable type:")

    for hashable in (m.Hashable, m.Hashable2):
        assert hashable(15) is not hashable(15)
        assert hashable(15) == hashable(15)

        assert hash(hashable(15)) == 15
        assert hash(hashable(15)) == hash(hashable(15))


def test_return_set_of_unhashable():
    with pytest.raises(TypeError) as excinfo:
        m.get_unhashable_HashMe_set()
    assert str(excinfo.value.__cause__).startswith("unhashable type:")



from helper import *

a = EUDVariable(0)


@EUDFunc
def f_func_with_side_effect(b):
    global a
    a += 1
    return b


@TestInstance
def test_eudscand():
    test_assert("EUDSCAnd w. (true, true)", EUDSCAnd()(Always())(Always())())
    test_assert("EUDSCAnd w. (true, false)", EUDNot(EUDSCAnd()(Always())(Never())()))
    test_assert("EUDSCAnd w. (false, false)", EUDNot(EUDSCAnd()(Never())(Never())()))

    # Check short-circuiting. f_func_with_side_effect should not be executed
    a << 0
    cv = EUDSCAnd()(Never())(f_func_with_side_effect(1))()
    test_assert("EUDSCAnd w. (false, true (side_effect)", EUDNot(cv))
    test_equality("EUDSCAnd w. (false, true (side_effect)", a, 0)

    # Check short-circuiting 2. f_func_with_side_effect should not be executed
    a << 0
    cv = EUDSCAnd()(f_func_with_side_effect(1))(f_func_with_side_effect(1))(
        f_func_with_side_effect(0)
    )(f_func_with_side_effect(1))()
    test_assert("EUDSCAnd w. (t, t, f, t)", EUDNot(cv))
    test_equality("EUDSCAnd w. (t, t, f, t)", a, 3)


@TestInstance
def test_eudscor():
    test_assert("EUDSCOr w. (true, true)", EUDSCOr()(Always())(Always())())
    test_assert("EUDSCOr w. (true, false)", EUDSCOr()(Always())(Never())())
    test_assert("EUDSCOr w. (false, false)", EUDNot(EUDSCOr()(Never())(Never())()))

    # Check short-circuiting. f_func_with_side_effect should not be executed
    a << 0
    cv = EUDSCOr()(Always())(f_func_with_side_effect(1))()
    test_assert("EUDSCOr w. (true, true (side_effect)", cv)
    test_equality("EUDSCOr w. (true, true (side_effect)", a, 0)

    # Check short-circuiting 2. f_func_with_side_effect should not be executed
    a << 0
    cv = EUDSCOr()(f_func_with_side_effect(0))(f_func_with_side_effect(0))(
        f_func_with_side_effect(1)
    )(f_func_with_side_effect(1))()
    test_assert("EUDSCOr w. (f, f, t, t)", cv)
    test_equality("EUDSCOr w. (f, f, t, t)", a, 3)



## NOTE: THIS FILE IS GENERATED BY EPSCRIPT! DO NOT MODITY
from eudplib import *
from eudplib.core.eudfunc import EUDTraceLog, EUDTracedFunc, EUDTracedTypedFunc, EUDTracedMethod, EUDTracedTypedMethod
from eudplib.epscript.helper import _RELIMP, _TYGV, _TYSV, _TYLV, _CGFW, _ARR, _VARR, _SRET, _SV, _ATTW, _ARRW, _ATTC, _ARRC, _L2V, _LSH, _ALL
# (Line 1) import .relimpCycle1;
# (Line 2) var B: Upgrade = "Terran Infantry Weapons";
try:
    relimpCycle1 = _RELIMP(".", "relimpCycle1")
except ImportError:
    from . import relimpCycle1
B = _TYGV([Upgrade], lambda: ["Terran Infantry Weapons"])



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from typing_extensions import Self

from .. import core as c
from .. import utils as ut
from ..ctrlstru import DoActions
from ..localize import _
from .strcommon import temp_string_id


class DBString(ut.ExprProxy):
    """Object for storing single modifiable string.

    Manipulating STR section is hard. DBString stores only one string, so that
    modifying its structure is substantially easier than modifying entire STR
    section. You can do anything you would do with normal string with DBString.
    """

    def __init__(self, content=None, *, _from=None) -> None:
        """Constructor for DBString

        :param content: Initial DBString content / capacity. Capacity of
            DBString is determined by size of this. If content is integer, then
            initial capacity and content of DBString will be set to
            content(int) and empty string.

        :type content: str, bytes, int
        """
        if _from is None:
            super().__init__(DBStringData(content))
        elif type(_from) in (str, bytes):
            super().__init__(DBStringData(_from))
        else:
            super().__init__(_from)

    def GetStringMemoryAddr(self):  # noqa: N802
        """Get memory address of DBString content.

        :returns: Memory address of DBString content.
        """
        return self

    def Display(self):  # noqa: N802
        strid = temp_string_id()
        DoActions(
            c.SetMemory(0x191943C8 + 4 * strid, c.SetTo, self - 0x191943C8),
            c.DisplayText(strid),
        )

    def Play(self):  # noqa: N802
        strid = temp_string_id()
        DoActions(
            c.SetMemory(0x191943C8 + 4 * strid, c.SetTo, self - 0x191943C8),
            c.PlayWAV(strid),
        )


class DBStringData(c.EUDObject):
    """Object containing DBString data"""

    def __new__(cls, *args, **kwargs) -> Self:
        return super().__new__(cls)

    def __init__(self, content) -> None:
        """Constructor for DBString

        :param content: Initial DBString content / capacity. Capacity of
            DBString is determined by size of this. If content is integer, then
            initial capacity and content of DBString will be set to
            content(int) and empty string.

        :type content: str, bytes, int
        """
        super().__init__()
        if isinstance(content, int):
            self.content = bytes(content)
        else:
            self.content = ut.u2utf8(content)
            if 0 in self.content:
                raise ut.EPError(
                    _("no nul bytes allowed in the middle of {}").format(
                        self.__class__
                    )
                )

    def GetDataSize(self):  # noqa: N802
        return len(self.content) + 1

    def WritePayload(self, pbuf):  # noqa: N802
        pbuf.WriteBytes(self.content)
        pbuf.WriteByte(0)



from helper import *


@TestInstance
def test_xvdefval():
    a = EUDXVariable(0, SetTo, 5, 19)
    test_equality(
        "XVariable with default values",
        [
            f_dwread(a.getDestAddr()),
            f_dwread(a._varact + 24),
            a,
            f_dwread(a.getMaskAddr()),
        ],
        [0, 0x072D0000, 5, 19],
    )


@TestInstance
def test_xvmixedtrg():
    # VMixed actions
    a = EUDXVariable(0, SetTo, 0)

    a << 0
    DoActions(SetDeaths(a, SetTo, EPD(a), 0))
    a << 1
    DoActions(SetDeaths(a, SetTo, f_mul(a, 30), 0))
    a << 2
    DoActions(SetDeaths(a, SetTo, a - 50, 0))
    DoActions(SetDeaths(3, SetTo, 123, a), SetDeaths(a, SetTo, a, a))

    test_assert(
        "XVariable mixed trigger test",
        [
            Deaths(0, Exactly, 0x3FE9D727, 0),  # EPD(0) == 0x3fe9d727
            Deaths(1, Exactly, 30, 0),
            Deaths(2, Exactly, -48, 0),
            Deaths(3, Exactly, 123, 2),
            Deaths(2, Exactly, 2, 2),
        ],
    )

    DoActions(a.SetNumber(15), a.SetMask(19))
    DoActions(SetDeaths(a, SetTo, 40 // a, a))
    d1 = EUDVariable()
    d1 << a
    d2, d3 = f_div(40, a)

    test_equality("XVariable flag trigger test", [d1, d2, d3], [3, 13, 1])

    DoActions(
        SetDeaths(AllPlayers, SetTo, 0, 0),
        SetDeaths(AllPlayers, SetTo, 0, 2),
        SetDeaths(AllPlayers, SetTo, 0, 3),
    )



import os
import sys

sys.path.insert(0, os.path.abspath("..\\"))

from eudplib import *


class Vector2D(EUDStruct):
    _fields_ = ["x", "y"]

    def constructor(self, x=None, y=None):
        self.x = x
        if y is not None:
            self.y = y

    def lengthsq(self):
        return self.x * self.x + self.y * self.y

    def length(self):
        return f_sqrt(self.lengthsq())

    def dot(self, rhs):
        return self.x * rhs.x + self.y * rhs.y

    # Operator support

    def add(self, rhs):
        t = self.copy()
        t += rhs
        return t

    def sub(self, rhs):
        t = self.copy()
        t -= rhs
        return t

    def mul(self, rhs):
        t = self.copy()
        t *= rhs
        return t

    def div(self, rhs):
        t = self.copy()
        t //= rhs
        return t

    def __iadd__(self, rhs):
        self.x += rhs.x
        self.y += rhs.y
        return self

    def __isub__(self, rhs):
        self.x -= rhs.x
        self.y -= rhs.y
        return self

    def __imul__(self, rhs):
        self.x *= rhs
        self.y *= rhs
        return self

    def __ifloordiv__(self, rhs):
        if EUDIf()(self.x >= 0x80000000):
            self.x = -(-self.x // rhs)
        if EUDElse()():
            self.x //= rhs
        EUDEndIf()

        if EUDIf()(self.y >= 0x80000000):
            self.y = -(-self.y // rhs)
        if EUDElse()():
            self.y //= rhs
        EUDEndIf()

        return self


radius = 100


class CircleObj(EUDStruct):
    _fields_ = [("pos", Vector2D), ("velocity", Vector2D)]

    def constructor(self, pos, velocity):
        self.pos = pos
        self.velocity = velocity

    def __init__(self, copyobj=None, *, pos=None, velocity=None):
        if copyobj is not None:
            super().__init__(_from=copyobj)

        else:
            if pos is None:
                pos = Vector2D()
            if velocity is None:
                velocity = Vector2D()
            super().__init__(pos, velocity)

    @EUDMethod
    def collides(self, rhs):
        self = CircleObj(self)
        rhs = CircleObj(rhs)

        distsq = rhs.pos.sub(self.pos).lengthsq()
        rsum = radius + radius
        rsumsq = rsum * rsum
        if EUDIf()(distsq <= rsumsq):
            EUDReturn(1)
        if EUDElse()():
            EUDReturn(0)
        EUDEndIf()

    def step(self):
        self.pos += self.velocity

    @EUDMethod
    def applyCollision(self, rhs):
        self = CircleObj(self)
        rhs = CircleObj(rhs)
        if EUDIf()(self.collides(rhs)):
            posdiff = rhs.pos.sub(self.pos)
            velavg = rhs.velocity.add(self.velocity).div(2)
            veldiff = rhs.velocity.sub(self.velocity)

            # Check if posdiff is in align with veldiff with dot
            aligning = posdiff.dot(veldiff)
            if EUDIf()(aligning >= 0x80000000):
                # Veldiff and posdiff are in opposite direction.
                # Meaning that they are doing collision

                # We create normal vector perpendicular with posdiff.
                posdiff_perpend = Vector2D()
                posdiff_perpend.setall([-posdiff.y, posdiff.x])
                posdiff_length = posdiff.length()

                # Divide velavg to perpendicular segment and parallel segment
                para_length = aligning
                perp_length = posdiff_perpend.dot(veldiff)

                # Invert para_length : collision
                para_length = -para_length

                # calculate veldiff back
                new_veldiff = (
                    posdiff.mul(para_length).add(posdiff_perpend.mul(perp_length))
                ).div(posdiff_length * posdiff_length)
                new_veldiff.copyto(veldiff)

                veldiff_d2 = veldiff.div(2)

                # Calculate self & rhs's velocity
                velavg.sub(veldiff_d2).copyto(self.velocity)
                velavg.add(veldiff_d2).copyto(rhs.velocity)
            EUDEndIf()
        EUDEndIf()


@EUDFunc
def main():
    a = CircleObj(pos=Vector2D(0, 0), velocity=Vector2D(10, 0))
    b = CircleObj(pos=Vector2D(1000, 0), velocity=Vector2D(-10, 0))

    if EUDInfLoop()():
        a.applyCollision(b)
        a.step()
        b.step()
        f_simpleprint(a.pos.x, a.pos.y, b.pos.x, b.pos.y)
        DoActions(SetMemory(0x6509A0, SetTo, 0))
        EUDDoEvents()
    EUDEndInfLoop()


LoadMap("outputmap/basemap/basemap.scx")
SaveMap("outputmap/collision.scx", main)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .. import core as c
from .. import ctrlstru as cs
from ..memio import f_setcurpl2cpcache
from ..memio.rwcommon import br1, br2, bs1, bw1
from ..scdata import CurrentPlayer
from ..utils import EPD


@c.EUDFunc
def f_strcpy(dst, src):
    """
    Strcpy equivalent in eudplib. Copy C-style string.

    :param dst: Destination address (Not EPD player)
    :param src: Source address (Not EPD player)

    :return: dst
    """
    b = c.EUDVariable()

    br1.seekoffset(src)
    bw1.seekoffset(dst)

    if cs.EUDInfLoop()():
        c.SetVariables(b, br1.readbyte())
        bw1.writebyte(b)
        cs.EUDBreakIf(b == 0)
    cs.EUDEndInfLoop()

    return dst


@c.EUDFunc
def f_strcmp(s1, s2):
    br1.seekoffset(s1)
    br2.seekoffset(s2)

    if cs.EUDInfLoop()():
        ch1 = br1.readbyte()
        ch2 = br2.readbyte()
        if cs.EUDIf()(ch1 == ch2):
            if cs.EUDIf()(ch1 == 0):
                c.EUDReturn(0)
            cs.EUDEndIf()
            cs.EUDContinue()
        if cs.EUDElse()():
            c.EUDReturn(ch1 - ch2)
        cs.EUDEndIf()
    cs.EUDEndInfLoop()


def f_strlen_epd(epd, subp=0, /, **kwargs):
    return _strlen_epd(epd, subp, **kwargs)


@c.EUDFunc
def _strlen_epd(epd, subp):
    ret = c.EUDVariable()
    b = [c.Forward() for _ in range(4)]
    jump = c.SetNextPtr(epd.GetVTable(), 0)
    loopend = c.Forward()
    for i, t in enumerate(b):
        c.RawTrigger(
            conditions=subp.ExactlyX(i, 3),
            actions=[
                c.SetMemory(jump + 20, c.SetTo, t),
                c.SetNextPtr(t, loopend),
            ],
        )
    c.RawTrigger(
        nextptr=epd.GetVTable(),
        actions=[jump, epd.SetDest(EPD(0x6509B0)), ret.SetNumber(0)],
    )
    if cs.EUDInfLoop()():
        b[0] << c.RawTrigger(
            nextptr=loopend,
            conditions=c.DeathsX(CurrentPlayer, c.AtLeast, 1, 0, 0xFF),
            actions=[
                c.SetNextPtr(b[0], b[1]),
                c.SetNextPtr(b[1], loopend),
                ret.AddNumber(1),
            ],
        )
        b[1] << c.RawTrigger(
            nextptr=loopend,
            conditions=c.DeathsX(CurrentPlayer, c.AtLeast, 1, 0, 0xFF00),
            actions=[
                c.SetNextPtr(b[1], b[2]),
                c.SetNextPtr(b[2], loopend),
                ret.AddNumber(1),
            ],
        )
        b[2] << c.RawTrigger(
            nextptr=loopend,
            conditions=c.DeathsX(CurrentPlayer, c.AtLeast, 1, 0, 0xFF0000),
            actions=[
                c.SetNextPtr(b[2], b[3]),
                c.SetNextPtr(b[3], loopend),
                ret.AddNumber(1),
            ],
        )
        b[3] << c.RawTrigger(
            nextptr=loopend,
            conditions=c.DeathsX(CurrentPlayer, c.AtLeast, 1, 0, 0xFF000000),
            actions=[
                c.SetNextPtr(b[3], b[0]),
                c.SetNextPtr(b[0], loopend),
                ret.AddNumber(1),
                c.SetMemory(0x6509B0, c.Add, 1),
            ],
        )
    cs.EUDEndInfLoop()
    loopend << c.NextTrigger()
    f_setcurpl2cpcache()
    c.EUDReturn(ret)


def f_strlen(src, /, **kwargs):
    epd, subp = c.f_div(src, 4)
    epd += -0x58A364 // 4
    return f_strlen_epd(epd, subp, **kwargs)


@c.EUDFunc
def f_strnstr(string, substring, count):
    bs1.seekoffset(string)
    br2.seekoffset(substring)
    dst = c.EUDVariable()
    dst << -1

    b = br2.readbyte()
    if cs.EUDIf()(b == 0):
        c.EUDReturn(string)
    cs.EUDEndIf()
    if cs.EUDWhile()(count >= 1):
        a = bs1.readbyte()
        cs.DoActions(dst.AddNumber(1), count.SubtractNumber(1))
        cs.EUDBreakIf(a == 0)
        cs.EUDContinueIfNot(a == b)
        oldoffset, oldsuboffset = c.EUDCreateVariables(2)
        c.VProc(
            [bs1._offset, bs1._suboffset],
            [
                bs1._offset.SetDest(oldoffset),
                bs1._suboffset.SetDest(oldsuboffset),
            ],
        )
        if cs.EUDInfLoop()():
            d = br2.readbyte()
            if cs.EUDIf()(d == 0):
                c.EUDReturn(string + dst)
            cs.EUDEndIf()
            cs.EUDBreakIfNot(bs1.readbyte() == d)
        cs.EUDEndInfLoop()
        c.VProc(
            [oldoffset, oldsuboffset],
            [
                oldoffset.SetDest(bs1._offset),
                oldsuboffset.SetDest(bs1._suboffset),
            ],
        )
        br2.seekoffset(substring + 1)
    cs.EUDEndWhile()
    c.EUDReturn(-1)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ...localize import _
from ...utils import EPError, u2b, u2utf8, unProxy
from .chktok import CHK
from .tblformat import TBL


class StringIdMap:
    def __init__(self) -> None:
        self._s2id: dict[bytes, int | None] = {}

    def add_item(self, string: str | bytes, strid: int) -> None:
        byte = u2b(unProxy(string))
        if byte in self._s2id:  # ambiguous string
            self._s2id[byte] = None
        else:
            self._s2id[byte] = strid

    def GetStringIndex(self, string: str | bytes) -> int:  # noqa: N802
        byte = u2utf8(unProxy(string))
        if byte not in self._s2id:
            byte = u2b(unProxy(string))
        if byte not in self._s2id:
            raise KeyError(string)
        retid = self._s2id.get(byte)
        if retid is None:
            raise KeyError(
                _(
                    "Ambiguous string; multiple entries have duplicated contents: {}"
                ).format(string)
            )
        return retid


strmap: TBL | None = None
strsection: str | None = None
unitmap: StringIdMap | None = None
locmap: StringIdMap | None = None
swmap: StringIdMap | None = None


def init_stringmaps(chkt: CHK) -> None:
    global strmap, strsection, unitmap, locmap, swmap

    unitmap = StringIdMap()
    locmap = StringIdMap()
    swmap = StringIdMap()
    init_chkt = (chkt, unitmap, locmap, swmap)

    try:
        strx = chkt.getsection("STRx")
    except KeyError:
        pass
    else:
        strsection = "STRx"
        strmap = TBL(strx, init_chkt, load_entry=4, save_entry=4)
        return
    strsection = "STRx"
    strmap = TBL(chkt.getsection("STR"), init_chkt, save_entry=4)


def GetLocationIndex(loc: str | bytes) -> int:  # noqa: N802
    if locmap is None:
        raise EPError(_("Must use LoadMap first"))
    return locmap.GetStringIndex(loc) + 1


def GetStringIndex(s: str | bytes) -> int:  # noqa: N802
    if strmap is None:
        raise EPError(_("Must use LoadMap first"))
    return strmap.GetStringIndex(s)


def GetSwitchIndex(s: str | bytes) -> int:  # noqa: N802
    if swmap is None:
        raise EPError(_("Must use LoadMap first"))
    return swmap.GetStringIndex(s)


def GetUnitIndex(u: str | bytes) -> int:  # noqa: N802
    if unitmap is None:
        raise EPError(_("Must use LoadMap first"))
    return unitmap.GetStringIndex(u)


def apply_string_map(chkt: CHK) -> None:
    if strsection is None or strmap is None:
        raise EPError(_("Must use LoadMap first"))
    chkt.setsection(strsection, strmap.save_tbl())


def ForceAddString(s: str | bytes) -> int:  # noqa: N802
    if strmap is None:
        raise EPError(_("Must use LoadMap first"))
    return strmap.ForceAddString(s) + 1


def get_string_map() -> TBL:
    if strmap is None:
        raise EPError(_("Must use LoadMap first"))
    return strmap


def get_string_section_name() -> str:
    if strsection is None:
        raise EPError(_("Must use LoadMap first"))
    return strsection



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from typing import Literal, TypedDict

from .. import core as c
from .. import utils as ut
from ..localize import _
from ..trigger import EUDBranch
from .basicstru import EUDJumpIf, EUDJumpIfNot
from .cshelper import CtrlStruOpener

"""
There are code duplication between EUDIf - EUDIfNot, EUDElseIf - EUDElseIfNot.
TODO : Remove code duplication if possible.
"""


def EUDIf() -> CtrlStruOpener:  # noqa: N802
    block = {
        "ifend": c.Forward(),
        "next_elseif": c.Forward(),
        "conditional": True,
    }
    ut.EUDCreateBlock("ifblock", block)

    def _footer(conditions, *, neg=False) -> Literal[True]:
        if neg:
            EUDJumpIf(conditions, block["next_elseif"])
        else:
            EUDJumpIfNot(conditions, block["next_elseif"])
        block["conditional"] = False
        return True

    return CtrlStruOpener(_footer)


def EUDIfNot() -> CtrlStruOpener:  # noqa: N802
    c = EUDIf()
    return CtrlStruOpener(lambda conditions: c(conditions, neg=True))


# -------


def EUDElseIf() -> CtrlStruOpener:  # noqa: N802
    def _header() -> None:
        block = ut.EUDPeekBlock("ifblock")[1]
        ut.ep_assert(
            block["next_elseif"] is not None,
            _("Cannot have EUDElseIf after EUDElse"),
        )

        # Finish previous if/elseif block
        c.SetNextTrigger(block["ifend"])

        block["next_elseif"] << c.NextTrigger()
        block["next_elseif"] = c.Forward()
        block["conditional"] = True

    def _footer(conditions, *, neg=False) -> Literal[True]:
        block = ut.EUDPeekBlock("ifblock")[1]
        if neg:
            EUDJumpIf(conditions, block["next_elseif"])
        else:
            EUDJumpIfNot(conditions, block["next_elseif"])
        block["conditional"] = False
        return True

    _header()
    return CtrlStruOpener(_footer)


def EUDElseIfNot() -> CtrlStruOpener:  # noqa: N802
    c = EUDElseIf()
    return CtrlStruOpener(lambda conditions: c(conditions, neg=True))


# -------


def EUDElse() -> CtrlStruOpener:  # noqa: N802
    def _footer() -> Literal[True]:
        block = ut.EUDPeekBlock("ifblock")[1]
        ut.ep_assert(
            block["next_elseif"] is not None,
            _("Cannot have EUDElse after EUDElse"),
        )

        # Finish previous if/elseif block
        c.SetNextTrigger(block["ifend"])
        block["next_elseif"] << c.NextTrigger()
        block["next_elseif"] = None
        return True

    return CtrlStruOpener(_footer)


def EUDEndIf() -> None:  # noqa: N802
    lb = ut.EUDPopBlock("ifblock")
    block = lb[1]

    # Finalize
    nei_fw = block["next_elseif"]
    if nei_fw:
        nei_fw << c.NextTrigger()

    block["ifend"] << c.NextTrigger()


# -------


def EUDExecuteOnce() -> CtrlStruOpener:  # noqa: N802
    def _header() -> None:
        class OnceBlock(TypedDict):
            blockstart: c.Forward
            blockend: c.Forward
            blockmode: bool
            conditional: bool

        block: OnceBlock = {
            "blockstart": c.Forward(),
            "blockend": c.Forward(),
            "blockmode": False,
            "conditional": True,
        }
        ut.EUDCreateBlock("executeonceblock", block)
        block["blockstart"] << c.NextTrigger()

    def _footer(conditions=None, *, neg=False) -> Literal[True]:
        block = ut.EUDPeekBlock("executeonceblock")[1]
        if conditions is not None:
            if neg:
                block["blockmode"] = False
                EUDJumpIf(conditions, block["blockend"])
            else:
                block["blockmode"] = True
                skip = [
                    c.SetNextPtr(block["blockstart"], block["blockend"]),
                    c.SetMemory(block["blockstart"] + 2376, c.SetTo, 8),
                ]
                ontrue = c.Forward()
                EUDBranch(conditions, ontrue, block["blockend"], _actions=skip)
                ontrue << c.NextTrigger()
        block["conditional"] = False
        return True

    _header()
    return CtrlStruOpener(_footer)


def EUDEndExecuteOnce() -> None:  # noqa: N802
    lb = ut.EUDPopBlock("executeonceblock")
    ut.ep_assert(lb[0] == "executeonceblock", _("Block start/end mismatch"))
    block = lb[1]
    if not block["blockmode"]:
        c.RawTrigger(
            actions=[
                c.SetNextPtr(block["blockstart"], block["blockend"]),
                c.SetMemory(block["blockstart"] + 2376, c.SetTo, 8),
            ]
        )
    block["blockend"] << c.NextTrigger()



from typing import TYPE_CHECKING, TypeAlias

from ...utils import ExprProxy
from .constenc import (
    TrgAllyStatus,
    TrgComparison,
    TrgCount,
    TrgModifier,
    TrgOrder,
    # TrgProperty
    TrgPropState,
    TrgResource,
    TrgScore,
    TrgSwitchAction,
    TrgSwitchState,
    UnitProperty,
    _Player,
)
from .consttype import Byte, Dword, Word
from .strdict import (
    DefaultAIScriptAtLocation,
    DefaultAIScriptWithoutLocation,
    DefaultButtonSet,
    DefaultFlingy,
    DefaultIcon,
    DefaultImage,
    DefaultIscript,
    DefaultPortrait,
    DefaultRank,
    DefaultSfxData,
    DefaultSprite,
    DefaultStatText,
    DefaultTech,
    DefaultUnit,
    DefaultUnitOrder,
    DefaultUpgrade,
    DefaultWeapon,
)

if TYPE_CHECKING:
    from ...scdata import (
        Flingy,
        Image,
        Sprite,
        Tech,
        TrgUnit,
        UnitOrder,
        Upgrade,
        Weapon,
    )

AllyStatus: TypeAlias = "TrgAllyStatus | Byte | ExprProxy[AllyStatus]"
Comparison: TypeAlias = "TrgComparison | Byte | ExprProxy[Comparison]"
Count: TypeAlias = "TrgCount | Byte | ExprProxy[Count]"
Modifier: TypeAlias = "TrgModifier | Byte | ExprProxy[Modifier]"
_Order: TypeAlias = "TrgOrder | Byte | ExprProxy[_Order]"
PropState: TypeAlias = "TrgPropState | Byte | ExprProxy[PropState]"
Resource: TypeAlias = "TrgResource | Byte | ExprProxy[Resource]"
_Score: TypeAlias = "TrgScore | Byte | ExprProxy[_Score]"
SwitchAction: TypeAlias = "TrgSwitchAction | Byte | ExprProxy[SwitchAction]"
SwitchState: TypeAlias = "TrgSwitchState | Byte | ExprProxy[SwitchState]"
_UnitProperty: TypeAlias = "UnitProperty | bytes | ExprProxy[_UnitProperty]"

AIScriptAtLocation: TypeAlias = (
    DefaultAIScriptAtLocation | str | bytes | Dword | ExprProxy["AIScriptAtLocation"]
)
"""4-characters AIScript are also allowed for input"""
AIScriptWithoutLocation: TypeAlias = """(
    DefaultAIScriptWithoutLocation
    | str
    | bytes
    | Dword
    | ExprProxy[AIScriptWithoutLocation]
)"""
ButtonSet: TypeAlias = "DefaultButtonSet | Word | ExprProxy[ButtonSet]"
_Icon: TypeAlias = "DefaultIcon | Word | ExprProxy[_Icon]"
_Iscript: TypeAlias = "DefaultIscript | Dword | ExprProxy[_Iscript]"
_Portrait: TypeAlias = "DefaultPortrait | Word | ExprProxy[_Portrait]"
Rank: TypeAlias = "DefaultRank | Byte | ExprProxy[Rank]"
SfxData: TypeAlias = "DefaultSfxData | str | Word | ExprProxy[SfxData]"
"""sfxdata.dat entries are case-insensitive"""
_StatText: TypeAlias = "DefaultStatText | Word | ExprProxy[_StatText]"

_Flingy: TypeAlias = "DefaultFlingy | Flingy | Byte | ExprProxy[_Flingy]"
_Image: TypeAlias = "DefaultImage | Image | Word | ExprProxy[_Image]"
Player: TypeAlias = "_Player | Dword | ExprProxy[Player]"
_Sprite: TypeAlias = "DefaultSprite | Sprite | Word | ExprProxy[_Sprite]"
_Tech: TypeAlias = "DefaultTech | Tech | Byte | ExprProxy[_Tech]"
Unit: TypeAlias = "DefaultUnit | TrgUnit | Word | ExprProxy[Unit]"
_UnitOrder: TypeAlias = "DefaultUnitOrder | UnitOrder | Byte | ExprProxy[_UnitOrder]"
_Upgrade: TypeAlias = "DefaultUpgrade | Upgrade | Byte | ExprProxy[_Upgrade]"
_Weapon: TypeAlias = "DefaultWeapon | Weapon | Byte | ExprProxy[_Weapon]"

Location: TypeAlias = "str | Byte | bytes | ExprProxy[Location]"
String: TypeAlias = "str | Word | bytes | ExprProxy[String]"
_Switch: TypeAlias = "str | Byte | bytes | ExprProxy[_Switch]"
"""Byte in Condition, Dword in Action"""



# Copyright 2019 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .. import utils as ut
from .memifgen import f_readgen_cp, f_readgen_epd

# CUnit
f_cunitread_epd = f_readgen_epd(0x3FFFF0, (0x400008, lambda x: x), _check_empty=True)
f_cunitread_cp = f_readgen_cp(0x3FFFF0, (0x400008, lambda x: x), _check_empty=True)
f_cunitepdread_epd = f_readgen_epd(
    0x3FFFF0,
    (0x400008, lambda x: x),
    (0x100002 - 0x58A364 // 4, lambda y: y // 4),
    _check_empty=True,
)
f_cunitepdread_cp = f_readgen_cp(
    0x3FFFF0,
    (0x400008, lambda x: x),
    (0x100002 - 0x58A364 // 4, lambda y: y // 4),
    _check_empty=True,
)
f_epdcunitread_epd = f_readgen_epd(
    0x3FFFF0,
    (0x100002 - 0x58A364 // 4, lambda y: y // 4),
    _check_empty=True,
)
f_epdcunitread_cp = f_readgen_cp(
    0x3FFFF0,
    (0x100002 - 0x58A364 // 4, lambda y: y // 4),
    _check_empty=True,
)
# CSprite
f_epdspriteread_epd = f_readgen_epd(
    0x1FFFC, (0x188000 - 0x58A364 // 4, lambda x: x // 4), _check_empty=True
)
f_epdspriteread_cp = f_readgen_cp(
    0x1FFFC, (0x188000 - 0x58A364 // 4, lambda x: x // 4), _check_empty=True
)
f_spriteepdread_epd = f_readgen_epd(
    0x1FFFC,
    (0x620000, lambda x: x),
    (0x188000 - 0x58A364 // 4, lambda y: y // 4),
    _check_empty=True,
)
f_spriteepdread_cp = f_readgen_cp(
    0x1FFFC,
    (0x620000, lambda x: x),
    (0x188000 - 0x58A364 // 4, lambda y: y // 4),
    _check_empty=True,
)
f_spriteread_epd = f_readgen_epd(0x1FFFC, (0x620000, lambda x: x), _check_empty=True)
f_spriteread_cp = f_readgen_cp(0x1FFFC, (0x620000, lambda x: x), _check_empty=True)


def _map_xy_mask():
    from ..core.mapdata import GetChkTokenized

    chkt = GetChkTokenized()
    dim = chkt.getsection("DIM")
    x, y = ut.b2i2(dim[0:2]), ut.b2i2(dim[2:4])
    x, y = (x - 1).bit_length(), (y - 1).bit_length()
    x = 2 ** (x + 5) - 1
    y = 2 ** (y + 5) - 1
    return x, y


def _posread_epd():
    f = getattr(_posread_epd, "f", None)
    if f is None:
        f = f_readgen_epd(
            (lambda x, y: x + 65536 * y)(*_map_xy_mask()),
            (0, lambda x: x if x <= 0xFFFF else 0),
            (0, lambda y: y >> 16),
        )
        _posread_epd.f = f
    return f


def f_posread_epd(epd, **kwargs):
    return _posread_epd()(epd, **kwargs)


def f_posread_cp(cpoffset, **kwargs):
    _rf = getattr(f_posread_cp, "_rf", None)
    if _rf is None:
        _rf = f_readgen_cp(
            (lambda x, y: x + 65536 * y)(*_map_xy_mask()),
            (0, lambda x: x if x <= 0xFFFF else 0),
            (0, lambda y: y >> 16),
        )
        f_posread_cp._rf = _rf
    return _rf(cpoffset, **kwargs)


def _mapxread_epd():
    f = getattr(_mapxread_epd, "f", None)
    if f is None:
        f = tuple(
            f_readgen_epd(_map_xy_mask()[0] << shift, (0, lambda x: x >> shift))
            for shift in (0, 16)
        )
        _mapxread_epd.f = f
    return f


def _mapyread_epd():
    f = getattr(_mapyread_epd, "f", None)
    if f is None:
        f = tuple(
            f_readgen_epd(_map_xy_mask()[1] << shift, (0, lambda x: x >> shift))
            for shift in (0, 16)
        )
        _mapyread_epd.f = f
    return f


def _playerread_epd():
    f = getattr(_playerread_epd, "f", None)
    if f is None:
        f = tuple(
            f_readgen_epd(0xF << shift, (0, lambda x: x >> shift))
            for shift in range(0, 32, 8)
        )
        _playerread_epd.f = f
    return f



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .eudf import (
    EUDFullFunc,
    EUDTracedFunc,
    EUDTracedTypedFunc,
    EUDXTypedFunc,
)
from .eudfmethod import (
    EUDTracedMethod,
    EUDTracedTypedMethod,
)
from .eudfuncn import EUDFuncN
from .trace.tracetool import EUDTraceLog, GetTraceStackDepth

__all__ = [
    "EUDFullFunc",
    "EUDTracedFunc",
    "EUDTracedTypedFunc",
    "EUDXTypedFunc",
    "EUDTracedMethod",
    "EUDTracedTypedMethod",
    "EUDFuncN",
    "EUDTraceLog",
    "GetTraceStackDepth",
]



# Copyright 2019 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ...localize import _
from ...utils import b2i2, ep_warn
from .chktok import CHK


def _fix_map_data(chkt: CHK) -> None:
    _fix_unit_map(chkt)
    _fix_stats_and_settings(chkt)
    _apply_remastered_chk(chkt)
    _fix_mtxm_0_0_null(chkt)


def _fix_unit_map(chkt: CHK) -> None:
    unit = bytearray(chkt.getsection("UNIT"))

    for i in range(0, len(unit), 36):
        # Disable flags for default value
        if unit[i + 17] == 100:
            unit[i + 14] &= ~2
        if unit[i + 18] == 100:
            unit[i + 14] &= ~4
        if unit[i + 24 : i + 26] == b"\0\0":
            unit[i + 14] &= ~32
        for k in range(5):
            if unit[i + 26] & (2**k) == 0:
                unit[i + 12] &= ~(2**k)

        # Remove disabled values
        # if unit[i + 14] & 1 == 0:
        #     unit[i + 16] = 0
        if unit[i + 14] & 2 == 0:
            unit[i + 17] = 0
        if unit[i + 14] & 4 == 0:
            unit[i + 18] = 0
        if unit[i + 14] & 8 == 0:
            unit[i + 19] = 0
        if unit[i + 14] & 16 == 0:
            unit[i + 20 : i + 24] = b"\0\0\0\0"
        if unit[i + 14] & 32 == 0:
            unit[i + 24 : i + 26] = b"\0\0"
        for k in range(5):
            if unit[i + 12] & (2**k) == 0:
                unit[i + 26] &= ~(2**k)

    chkt.setsection("UNIT", unit)


def _fix_stats_and_settings(chkt: CHK) -> None:
    sections = (
        ("UNIx", 228, (4, 2, 1, 2, 2, 2, 2)),
        ("UPGx", 61, (2, 2, 2, 2, 2, 2)),
        ("TECx", 44, (2, 2, 2, 2)),
    )
    for name, count, settings in sections:
        data = bytearray(chkt.getsection(name))

        for i in range(count):
            if data[i] == 0:
                continue
            offset = count
            if name == "UPGx":
                offset += 1
            for setting in settings:
                for k in range(setting):
                    data[offset + setting * i + k] = 0
                offset += setting * count

        chkt.setsection(name, data)


def _apply_remastered_chk(chkt: CHK) -> None:
    chkt.setsection("VER ", b"\xCE\0")


def _fix_mtxm_0_0_null(chkt: CHK) -> None:
    mtxm = bytearray(chkt.getsection("MTXM"))

    null_tiles = []
    for i in range(0, len(mtxm), 2):
        if mtxm[i : i + 2] == b"\0\0":
            mtxm[i : i + 2] = b"\x01\0"
            null_tiles.append(i // 2)

    if null_tiles:
        dim = chkt.getsection("DIM")
        width = b2i2(dim, 0)

        print(
            "Null tiles at:",
            ", ".join(map(lambda x: str(divmod(x, width)[::-1]), null_tiles)),
        )
        ep_warn(
            _("[Warning] Input map has 0000.00 null tiles")
            + "\n"
            + _("Replaced them to 0000.01, because they cause desync.")
        )

    chkt.setsection("MTXM", mtxm)



from helper import *


@EUDFunc
def f1(a):
    ret = EUDVariable(0)
    if EUDIf()(a == 0):
        ret << 1234
    if EUDElse()():
        ret << 5678
    EUDEndIf()

    return ret


@EUDFunc
def f2(a):
    if EUDIf()(a == 0):
        EUDReturn(1234)
    if EUDElseIf()(a == 1):
        EUDReturn(5678)
    EUDEndIf()
    return 9012


@TestInstance
def test_multret():
    # a, b = f1(0), f1(1)
    a, b = EUDCreateVariables(2)
    f1(0, ret=[a])
    f1(1, ret=[b])
    c, d, e = f2(0), f2(1), f2(2)
    test_assert(
        "Multiple EUDReturn test",
        [a == 1234, b == 5678, c == 1234, d == 5678, e == 9012],
    )



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .. import core as c
from .. import ctrlstru as cs
from ..memio import (
    EUDByteWriter,
    f_dwread_epd,
    f_memcpy,
    f_setcurpl2cpcache,
)
from ..scdata import CurrentPlayer, TrgUnit
from ..utils import EPD, ep_assert

_PROV_MAXBUFFER = 0x57F0D8
_cmdqlen = c.EUDVariable()
_CMDQLEN = 0x654AA0
_len_cache = c.Memory(_CMDQLEN, c.Exactly, 0)


@c.EUDFunc
def _update_cmdqlen():
    if cs.EUDIfNot()(_len_cache):
        f_dwread_epd(EPD(0x654AA0), ret=[_cmdqlen])
        c.VProc(_cmdqlen, _cmdqlen.QueueAssignTo(EPD(_len_cache) + 2))
    cs.EUDEndIf()


def _get_cmdqlen():
    _update_cmdqlen()
    return _cmdqlen


def _set_cmdqlen(new_len):
    ep_assert(c.IsEUDVariable(new_len), "new command queue length can't be constant")
    c.VProc(
        [new_len, _cmdqlen],
        [
            new_len.QueueAssignTo(_cmdqlen),
            _cmdqlen.QueueAssignTo(EPD(_CMDQLEN)),
        ],
    )
    c.VProc(_cmdqlen, _cmdqlen.SetDest(EPD(_len_cache) + 2))


class QueueGameCommandHelper:
    def __init__(self, size):
        self.size = size

    def __enter__(self):
        self.cmdqlen = _get_cmdqlen()
        self.new_len = _cmdqlen + self.size
        if cs.EUDIfNot()(c.Memory(_PROV_MAXBUFFER, c.AtMost, self.new_len + 1)):
            return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        if exception_type is None:
            _set_cmdqlen(self.new_len)
            cs.EUDEndIf()


@c.EUDFunc
def QueueGameCommand(data, size):  # noqa: N802
    """Queue game command to packet queue.

    Starcraft periodically broadcasts game packets to other player. Game
    packets are stored to queue, and this function add data to that queue, so
    that SC can broadcast it.

    :param data: Data to put in queue
    :param size: Size of data

    .. note::
        If packet queue is full, this function fails. This behavior is silent
        without any warning or error, since this behavior shouldn't happen in
        common situations. So **Don't use this function too much in a frame.**

    """
    with QueueGameCommandHelper(size) as qgc:
        f_memcpy(0x654880 + qgc.cmdqlen, data, qgc.size)


bw = EUDByteWriter()


@c.EUDFunc
def _qgc_alphaids(packet_id, n, arr_epd):
    """
    == 0x0B - Select Delta Del ==
    == 0x0A - Select Delta Add ==
    == 0x09 - Select Units ==
    """
    with QueueGameCommandHelper(2 + 2 * n) as qgc:
        bw.seekoffset(0x654880 + qgc.cmdqlen)
        bw.writebyte(packet_id)
        bw.writebyte(n)

        cp2arr = c.SetMemory(0x6509B0, c.SetTo, 0)
        cp2arr_quantity = EPD(cp2arr) + 5
        c.SetVariables(cp2arr_quantity, arr_epd)
        if cs.EUDWhileNot()(n.Exactly(0)):
            b0, b1 = c.EUDCreateVariables(2)
            restore_arr = c.SetDeaths(CurrentPlayer, c.Add, 0, 0)
            cp2uniqueness_identifier = c.SetMemory(0x6509B0, c.SetTo, 0)
            cp2ui_quantity = EPD(cp2uniqueness_identifier) + 5
            cunit_initval = EPD(0x59CCA8) + 0xA5 // 4
            c.RawTrigger(
                actions=[
                    cp2arr,
                    b0.SetNumber(1),
                    b1.SetNumber(0),
                    c.SetMemory(restore_arr + 20, c.SetTo, 0),
                    c.SetMemoryEPD(cp2ui_quantity, c.SetTo, cunit_initval),
                ]
            )
            for i in range(10, -1, -1):
                ptr = 0x59CCA8 + (336 << i)
                c.RawTrigger(
                    conditions=c.Deaths(CurrentPlayer, c.AtLeast, ptr, 0),
                    actions=[
                        c.SetDeaths(CurrentPlayer, c.Subtract, 336 << i, 0),
                        c.SetMemory(restore_arr + 20, c.Add, 336 << i),
                        c.SetMemoryEPD(cp2ui_quantity, c.Add, 84 << i),
                        b0.AddNumber(1 << i)
                        if i < 8
                        else b1.AddNumber(1 << (i - 8)),
                    ],
                )
            c.RawTrigger(
                conditions=b0.AtLeast(256),
                actions=[
                    b0.SubtractNumber(256),
                    b1.AddNumber(1),
                ],
            )
            c.RawTrigger(actions=[restore_arr, cp2uniqueness_identifier])
            for i in range(8):
                m = 1 << (i + 8)
                c.RawTrigger(
                    conditions=c.DeathsX(CurrentPlayer, c.AtLeast, 1, 0, m),
                    actions=b1.AddNumber(1 << (i + 3)),
                )
            bw.writebyte(b0)
            bw.writebyte(b1)
            c.SetVariables([cp2arr_quantity, n], [1, 1], [c.Add, c.Subtract])
        cs.EUDEndWhile()
        f_setcurpl2cpcache()


def QueueGameCommand_Select(n, ptr_arr):  # noqa: N802
    """
    == 0x09 - Select Units ==
    {{{
        BYTE bCount;
        WORD unitID[bCount];
    }}}
    """
    if isinstance(ptr_arr, int) or c.IsEUDVariable(ptr_arr):
        _qgc_alphaids(0x09, n, EPD(ptr_arr))
    elif c.IsConstExpr(ptr_arr) and ptr_arr._is_epd():
        _qgc_alphaids(0x09, n, ptr_arr)
    else:
        _qgc_alphaids(0x09, n, EPD(ptr_arr))


def QueueGameCommand_AddSelect(n, ptr_arr):  # noqa: N802
    """
    == 0x0A - Select Delta Add ==
    {{{
        BYTE bCount;
        WORD unitID[bCount];
    }}}
    """
    if isinstance(ptr_arr, int) or c.IsEUDVariable(ptr_arr):
        _qgc_alphaids(0x0A, n, EPD(ptr_arr))
    elif c.IsConstExpr(ptr_arr) and ptr_arr._is_epd():
        _qgc_alphaids(0x0A, n, ptr_arr)
    else:
        _qgc_alphaids(0x0A, n, EPD(ptr_arr))


def QueueGameCommand_RemoveSelect(n, ptr_arr):  # noqa: N802
    """
    == 0x0B - Select Delta Del ==
    {{{
        BYTE bCount;
        WORD unitID[bCount];
    }}}
    """
    if isinstance(ptr_arr, int) or c.IsEUDVariable(ptr_arr):
        _qgc_alphaids(0x0B, n, EPD(ptr_arr))
    elif c.IsConstExpr(ptr_arr) and ptr_arr._is_epd():
        _qgc_alphaids(0x0B, n, ptr_arr)
    else:
        _qgc_alphaids(0x0B, n, EPD(ptr_arr))


@c.EUDFunc
def QueueGameCommand_RightClick(xy):  # noqa: N802
    """Queue right click action.

    :param xy: (y * 65536) + x, where (x, y) is coordinate for right click.
    """
    right_click_command = c.Db(b"...\x14XXYY\0\0\xE4\0\x00")
    c.SetVariables(EPD(right_click_command + 4), xy)
    QueueGameCommand(right_click_command + 3, 10)


@c.EUDFunc
def QueueGameCommand_QueuedRightClick(xy):  # noqa: N802
    """Queue right click action.

    :param xy: (y * 65536) + x, where (x, y) is coordinate for right click.
    """
    queued_right_click_command = c.Db(b"...\x14XXYY\0\0\xE4\0\x01")
    c.SetVariables(EPD(queued_right_click_command + 4), xy)
    QueueGameCommand(queued_right_click_command + 3, 10)


@c.EUDFunc
def QueueGameCommand_MinimapPing(xy):  # noqa: N802
    """Queue minimap ping action.

    :param xy: (y * 65536) + x, where (x, y) is coordinate for minimap ping.
    """
    minimap_ping_command = c.Db(b"...\x58XXYY")
    c.SetVariables(EPD(minimap_ping_command + 4), xy)
    QueueGameCommand(minimap_ping_command + 3, 5)


@c.EUDTypedFunc([TrgUnit])
def QueueGameCommand_TrainUnit(unit):  # noqa: N802
    train_unit_command = c.Db(b"...\x1FUU..")
    c.SetVariables(EPD(train_unit_command + 4), unit)
    QueueGameCommand(train_unit_command + 3, 3)


@c.EUDFunc
def QueueGameCommand_PauseGame():  # noqa: N802
    pause_game_command = c.Db(b"\x10")
    QueueGameCommand(pause_game_command, 1)


@c.EUDFunc
def QueueGameCommand_ResumeGame():  # noqa: N802
    resume_game_command = c.Db(b"\x11")
    QueueGameCommand(resume_game_command, 1)


@c.EUDFunc
def QueueGameCommand_RestartGame():  # noqa: N802
    restart_game_command = c.Db(b"\x08")
    QueueGameCommand(restart_game_command, 1)


@c.EUDFunc
def QueueGameCommand_MergeDarkArchon():  # noqa: N802
    merge_dark_archon_command = c.Db(b"\x5A")
    QueueGameCommand(merge_dark_archon_command, 1)


@c.EUDFunc
def QueueGameCommand_MergeArchon():  # noqa: N802
    merge_archon_command = c.Db(b"\x2A")
    QueueGameCommand(merge_archon_command, 1)


@c.EUDFunc
def QueueGameCommand_UseCheat(flags):  # noqa: N802
    use_cheat_command = c.Db(b"...\x12CCCC")
    c.SetVariables(EPD(use_cheat_command + 4), flags)
    QueueGameCommand(use_cheat_command + 3, 5)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from collections.abc import Iterator, Sequence


def b2i1(b: Sequence[int], index: int = 0) -> int:
    return b[index]


def b2i2(b: Sequence[int], index: int = 0) -> int:
    return b[index] | (b[index + 1] << 8)


def b2i4(b: Sequence[int], index: int = 0) -> int:
    return (
        b[index] | (b[index + 1] << 8) | (b[index + 2] << 16) | (b[index + 3] << 24)
    )


def i2b1(i: int) -> bytes:
    return bytes((i & 0xFF,))


def i2b2(i: int) -> bytes:
    i &= 0xFFFF
    return bytes((i & 0xFF, (i >> 8) & 0xFF))


def i2b4(i: int) -> bytes:
    i &= 0xFFFFFFFF
    return bytes((i & 0xFF, (i >> 8) & 0xFF, (i >> 16) & 0xFF, (i >> 24) & 0xFF))


def bits(n: int) -> Iterator[int]:
    n = n & 0xFFFFFFFF
    while n:
        b = n & (~n + 1)
        yield b
        n ^= b
